# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..application import W_APP_BALANCE
from ..defines import ErrorDef
from ..variable import Var, strjoin


# TODO: instead of where, we should pass a context object that context
# could include a where, and also a list of classes where we can store
# significant info like "label" when there is a goto/gosub that jumps to
# one.


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_APP_ARG_FEW(ErrorDef):
        message = 'too few arguments for app {app!r}, minimum is {min_args}'

    class E_APP_ARG_MANY(ErrorDef):
        message = 'too many arguments for app {app!r}, maximum is {max_args}'

    class E_APP_ARG_BADOPT(ErrorDef):
        message = ('unrecognised options {opts!r} in arg {argno} '
                   'for app {app!r}')

    class E_APP_ARG_DUPEOPT(ErrorDef):
        message = 'duplicate options {opts!r} in arg {argno} for app {app!r}'

    class E_APP_ARG_IFSTYLE(ErrorDef):
        message = ("{app!r} takes the form <cond>?<iftrue>[:<iffalse>] "
                   "but data is '{data}', cond '{cond}', args '''{args}'''")

    class E_APP_ARG_PIPEDELIM(ErrorDef):
        message = ('the application delimiter is now the comma, not '
                   'the pipe; see app {app!r} and data {data!r}')


class AppArg(object):
    def __init__(self, name):
        self.name = name
        # The following are set afterwards.
        self.argno = None
        self.app = None

    def validate(self, arg, where):
        pass


class AppOptions(AppArg):
    def __init__(self, options):
        super().__init__('options')
        self.options = options

    def validate(self, arg, where):
        str_options = [i for i in arg if isinstance(i, str)]
        bad_options = [i for i in str_options if i not in self.options]
        if bad_options:
            E_APP_ARG_BADOPT(where, argno=self.argno, app=self.app,
                             opts=''.join(bad_options))
        if len(str_options) != len(set(str_options)):
            E_APP_ARG_DUPEOPT(where, argno=self.argno, app=self.app,
                              opts=arg)


class AppBase(object):
    @property
    def name(self):
        return self.__class__.__name__

    @property
    def module(self):
        return self.__module__.rsplit('.', 1)[-1]

    def __call__(self, data, where):
        try:
            self.check_balance(data)
        except ValueError:
            W_APP_BALANCE(where, data=str(data))

    @staticmethod
    def check_balance(data):
        """
        This function is a simple check that looks sane, but is not
        implemented like this in Asterisk. It may warn you of badly
        placed delimiters, but you shouldn't use this to interpret
        variables.

        For that, see separate_args which is modeled after the function
        as implemented in Asterisk.
        """
        arr = ['X']
        for char in data:
            if isinstance(char, Var):
                # It's possible that we're looping over a Var variable.
                # In that case we'll have to assume the Var itself
                # contains no separators that we might be interested in.
                # E.g. we expect you do *not* do this:
                # ``Set(var=mailbox@context,s)``
                # ``VoiceMail(${var})``
                # But we expect you to do this:
                # ``Set(mailbox=mailbox@context)``
                # ``VoiceMail(${mailbox},s)``
                pass  # skip char
            elif char == '"':
                if arr[-1] == '"':
                    arr.pop()
                elif arr[-1] == "'":
                    pass
                else:
                    arr.append('"')
            elif char == "'":
                if arr[-1] == "'":
                    arr.pop()
                elif arr[-1] == '"':
                    pass
                else:
                    arr.append("'")
            elif char in '({[':
                if arr[-1] in '\'"':
                    pass
                else:
                    arr.append(char)
            elif char in ')}]':
                left = '({['[')}]'.index(char)]
                if arr[-1] in '\'"':
                    pass
                else:
                    if arr[-1] == left:
                        arr.pop()
                    else:
                        raise ValueError(''.join(arr[1:]))
        if arr != ['X']:
            raise ValueError(''.join(arr[1:]))

    @staticmethod
    def separate_args(data, delimiter=',', remove_quotes_backslashes=True):
        # #define AST_STANDARD_APP_ARGS(...) => separate_args(',', True)
        # #define AST_STANDARD_RAW_ARGS(...) => separate_args(',', False)
        # SOURCE: main/app.c -- __ast_app_separate_args()
        #
        # TODO: we should separate args using a more sensible approach as
        # well, so we can warn on inconsistencies.
        brackets = 0
        parens = 0
        quotes = False
        skipnext = False

        ret = [[]]
        start = 0
        for i, char in enumerate(data):
            if isinstance(char, Var):
                # It's possible that we're looping over a Var variable.
                # In that case we'll have to assume the Var itself
                # contains no separators that we might be interested in.
                # E.g. we expect you do *not* do this:
                # ``Set(var=mailbox@context,s)``
                # ``VoiceMail(${var})``
                # But we expect you to do this:
                # ``Set(mailbox=mailbox@context)``
                # ``VoiceMail(${mailbox},s)``
                # Skip it?
                # #ret[-1].extend(data[start:i])  # skip char
                # #start = i + 1
                # Don't skip it, we want the returned value.
                pass
            elif skipnext:
                skipnext = False
            elif char == '[':
                brackets += 1
            elif char == ']':
                if brackets:
                    brackets -= 1
            elif char == '(':
                parens += 1
            elif char == ')':
                if parens:
                    parens -= 1
            elif char == '"' and delimiter != '"':
                quotes = not quotes
                if remove_quotes_backslashes:
                    ret[-1].extend(data[start:i])
                    start = i + 1
            elif char == '\\':
                if remove_quotes_backslashes:
                    ret[-1].extend(data[start:i])
                    start = i + 1
                skipnext = True
            elif char == delimiter and not (brackets or parens or quotes):
                ret[-1].extend(data[start:i])
                start = i + 1
                ret.append([])  # start on next arg

        # Append leftover args.
        ret[-1].extend(data[start:])

        # Squash args.
        squashed = []
        for letters in ret:
            letters = list(strjoin(letters))  # join sequences of strings
            if len(letters) == 1:
                squashed.append(letters[0])
            else:
                squashed.append(Var.join(letters))

        return squashed


class DelimitedArgsMixin(object):
    def __init__(self, arg_delimiter=',', arg_raw=False, **kwargs):
        self._arg_delimiter = arg_delimiter
        self._arg_raw = arg_raw

    def split_args(self, data, where):
        return self.separate_args(
            data, delimiter=self._arg_delimiter,
            remove_quotes_backslashes=self._arg_raw)

    def __call__(self, data, where):
        return self.split_args(data, where)


class MinMaxArgsMixin(object):
    def __init__(self, min_args=None, max_args=None, **kwargs):
        super().__init__(**kwargs)
        assert (min_args is None or max_args is None or
                min_args <= max_args)
        self._min_args = min_args
        self._max_args = max_args

    def split_args(self, data, where):
        args = super().split_args(data, where)

        if self._min_args is not None and (
                (len(args) < self._min_args) or
                (self._min_args == 1 and len(args) == 1 and not args[0])):
            E_APP_ARG_FEW(where, app=self.name, min_args=self._min_args)
        elif self._max_args is not None and (
                (len(args) > self._max_args)):
            E_APP_ARG_MANY(where, app=self.name, max_args=self._max_args)

        return args


class AppArgsMixin(MinMaxArgsMixin):
    def __init__(self, args=None, **kwargs):
        assert args, 'AppArgsMixin wants some args for {!r}'.format(
            self.__class__.__name__)

        super().__init__(max_args=len(args), **kwargs)

        # Set the arguments and pimp them a bit with more info.
        self._args = args
        for i, arg in enumerate(self._args):
            arg.argno = i + 1
            arg.app = self.name

    def split_args(self, data, where):
        args = super().split_args(data, where)

        for i, arg in enumerate(args):
            if len(self._args) > i:
                self._args[i].validate(arg, where)

        return args


class NoPipeDelimiterMixin(object):
    # TODO: we should not use this for the System() app,
    # nor for the SHELL() function,
    # and it would be improved if we checked for surrounding quotes
    # which clearly show that the user isn't doing anything unintended.

    def split_args(self, data, where):
        args = super().split_args(data, where)
        if len(args) == 1 and '|' in args[0]:
            E_APP_ARG_PIPEDELIM(where, app=self.name, data=data)

        return args


class App(NoPipeDelimiterMixin, AppArgsMixin, DelimitedArgsMixin, AppBase):
    """
    The App takes an optional args=[AppArg(...), ...] and max_args=INT,
    and checks the values if available.
    """
    pass


class IfStyleApp(DelimitedArgsMixin, AppBase):
    """
    The IfStyleApp takes the form <cond>?<iftrue>:<iffalse>.
    """
    def __init__(self, **kwargs):
        super().__init__(arg_delimiter='?', arg_raw=True, **kwargs)

    def split_args(self, data, where):
        args = super().split_args(data, where)
        if len(args) == 1:
            # Attempt to separate the arguments the old fashioned way.
            args = self.separate_args(
                data, ',', remove_quotes_backslashes=False)
            E_APP_ARG_IFSTYLE(where, app=self.name, data=data, cond=args[0],
                              args=args[1:])
            return args[0], args[1:], None
        elif len(args) > 2:
            # This is no good either, with too many results.
            E_APP_ARG_IFSTYLE(where, app=self.name, data=data, cond=args[0],
                              args=args[1:])
            return args

        # BEWARE: GosubIf and ExecIf use this method, while GotoIf
        # uses a simpler split. In the majority of cases the result
        # would be the same (barring crazy backslash usage).
        cond = args[0]
        actions = self.separate_args(
            args[1], delimiter=':', remove_quotes_backslashes=False)
        if len(actions) == 1:
            iftrue, iffalse = actions[0], None
        elif len(actions) == 2:
            iftrue, iffalse = actions[0], actions[1]
        else:
            return None

        return cond, iftrue, iffalse



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re

from .config import (
    E_CONF_KEY_DUPE, E_CONF_KEY_INVALID, W_CONF_KEY_LATE,
    ConfigAggregator, Context)
from .defines import ErrorDef, WarningDef
from .func.base import FuncBase
from .varfun import FuncLoader


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_ODBC_NOSQL(ErrorDef):
        message = ('a readsql and/or writesql (and insertsql) statement is '
                   'required')

    class W_ODBC_BADTOKENS(WarningDef):
        message = ('odbc function {name!r} should be uppercase '
                   'with A-Z0-9_ only')

    class W_ODBC_LEGACY(WarningDef):
        message = ('read and write are legacy, use readsql and writesql '
                   '(and insertsql) instead')

    class W_ODBC_TAILSEMI(WarningDef):
        message = 'tailing semicolon is unnecessary'


class FuncOdbcFunc(FuncBase):
    @property
    def module(self):
        # This would be correct automatically, since this module is
        # called func_odbc, but it doesn't hurt to be explicit.
        return 'func_odbc'


class OdbcFunction(Context):
    ORDER = ('prefix', 'synopsis', 'dsn', 'escapecommas', 'mode', 'rowlimit',
             # 'insertsql' is used when the writesql update returned 0
             # updated rows.
             'readhandle', 'readsql', 'writehandle', 'writesql', 'insertsql',
             'read', 'write')  # legacy

    def get_name(self):
        return self.name

    def get_function_name(self):
        return '{}_{}'.format(self.get_prefix(), self.get_name())

    def get_prefix(self):
        return self.get_variable('prefix') or 'ODBC'

    def get_variable(self, variable):
        for var in self._varsets:
            if var.variable == variable:
                return var.value
        return None

    def has_variable(self, variable):
        return self.get_variable(variable) is not None

    def add(self, varset):
        if self.check_order(self._varsets, varset):

            # Check for trailing semicolons.
            if varset.variable in (
                    'readsql', 'writesql', 'insertsql', 'read', 'write'):
                if varset.variable in ('read', 'write'):
                    W_ODBC_LEGACY(varset.where)
                if varset.value.endswith(';'):
                    W_ODBC_TAILSEMI(varset.where)
            # TODO:
            # - check for usage/missing ARGn/VALn (only VALn in
            #   write/insert)
            # - properly formatted ${SQL_ESC(${ARG4})}
            # - prefix should be uppercase
            # - check SQL syntax?

            return super(OdbcFunction, self).add(varset)

    def update(self, context):
        raise NotImplementedError('when? how? what?')

    def check_order(self, set_, new):
        try:
            new_pos = self.ORDER.index(new.variable)
        except ValueError:
            E_CONF_KEY_INVALID(new.where, key=new.variable)
            return False

        if set_:
            used_poses = [self.ORDER.index(i.variable) for i in set_]
            if new_pos in used_poses:
                # func_odbc.c uses ast_variable_retrieve (checked on jan
                # 2016 in Asterisk 13), so it uses the first key, not
                # the last.
                previous = [i for i in set_ if i.variable == new.variable][0]
                E_CONF_KEY_DUPE(new.where, key=new.variable, previous=previous)
                return False

            if new_pos < used_poses[-1]:
                # We expected this key sooner.
                W_CONF_KEY_LATE(new.where, key=new.variable)

        return True

    def finalize(self):
        # TODO:
        # - Check that we have not both read and readsql (and friends)
        # - Check that there is a dsn or a *handle for every *sql.
        # - Check existence of synopsis/syntax
        # - Note: syntax=(empty) does not work, use syntax=void or
        #   syntax=arg1,arg2 (syntax check should match comma's to
        #   ARGn count)

        if not self.finalize_check_any_read_write():
            return
        self.finalize_register()

    def finalize_check_any_read_write(self):
        if not (self.has_variable('read') or self.has_variable('write') or
                self.has_variable('readsql') or self.has_variable('writesql')):
            E_ODBC_NOSQL(self.where)
            return False
        return True

    def finalize_register(self):
        # Register the custom function.
        odbc_function_class = type(
            self.get_function_name(), (FuncOdbcFunc,), {})
        FuncLoader().register(odbc_function_class())


class FuncOdbcAggregator(ConfigAggregator):
    """
    TODO: This should instead be split up into a simple config parser
    which also handles inheritance and a really short func_odbc
    aggregator.
    """
    # Allow A_B_C and ABC, but not _A or B_ or AB__C.
    FUNC_NAME_RE = re.compile(r'^[A-Z0-9](_?[A-Z0-9])*$')

    def on_yield(self):
        for function in super(FuncOdbcAggregator, self).on_yield():
            function.finalize()
            yield function

        # TODO: - Check for dupe functions/contexts.  jan2016 check in
        # pbx.c says that duplicate functions do not get registered =>
        # first one wins

    def on_context(self, context):
        if not self.FUNC_NAME_RE.match(context.name):
            W_ODBC_BADTOKENS(context.where, name=context.name)

        context = OdbcFunction.from_context(context)
        return super(FuncOdbcAggregator, self).on_context(context)

    def on_varset(self, varset):
        assert not varset.arrow  # W_ARROW
        return super(FuncOdbcAggregator, self).on_varset(varset)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import FuncBase


class ARRAY(FuncBase):
    pass


class CSV_QUOTE(FuncBase):
    pass


class EVAL(FuncBase):
    pass


class FIELDNUM(FuncBase):
    pass


class FIELDQTY(FuncBase):
    pass


class FILTER(FuncBase):
    pass


class HASH(FuncBase):
    pass


class HASHKEYS(FuncBase):
    pass


class KEYPADHASH(FuncBase):
    pass


class LEN(FuncBase):
    pass


class LISTFILTER(FuncBase):
    pass


class PASSTHRU(FuncBase):
    pass


class POP(FuncBase):
    pass


class PUSH(FuncBase):
    pass


class QUOTE(FuncBase):
    pass


class REGEX(FuncBase):
    pass


class REPLACE(FuncBase):
    pass


class SHIFT(FuncBase):
    pass


class STRFTIME(FuncBase):
    pass


class STRPTIME(FuncBase):
    pass


class STRREPLACE(FuncBase):
    pass


class TOLOWER(FuncBase):
    pass


class TOUPPER(FuncBase):
    pass


class UNSHIFT(FuncBase):
    pass


def register(func_loader):
    for func in (
            ARRAY,
            CSV_QUOTE,
            EVAL,
            FIELDNUM,
            FIELDQTY,
            FILTER,
            HASH,
            HASHKEYS,
            KEYPADHASH,
            LEN,
            LISTFILTER,
            PASSTHRU,
            POP,
            PUSH,
            QUOTE,
            REGEX,
            REPLACE,
            SHIFT,
            STRFTIME,
            STRPTIME,
            STRREPLACE,
            TOLOWER,
            TOUPPER,
            UNSHIFT,
            ):
        func_loader.register(func())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from collections import defaultdict
from io import BytesIO
from unittest import (
    TestCase, TextTestResult, TextTestRunner, main as orig_main)

from .defines import MessageDefManager


__all__ = ('ALintTestCase', 'NamedBytesIO', 'ignoreLinted')


class ALintTestCase(TestCase):
    def setUp(self):
        MessageDefManager.muted = True  # could do this on runner startup..
        self.linted_counts = defaultdict(int)

    def tearDown(self):
        if self.last_test_was_a_success():
            self.assertLinted({})
        MessageDefManager.reset()

    def last_test_was_a_success(self):
        """
        Hope that unittest for Python3.4 doesn't change.
        """
        # If the test was "supposed to fail", we still want it reported
        # as a failure.
        if self._outcome.expectedFailure:
            return False

        # The error list holds accumulated status reports (not all of
        # them errors).
        last_error = self._outcome.errors[-1]
        test_method, errors = last_error
        return not bool(errors)

    def assertLinted(self, expected_counts):
        raised = MessageDefManager.raised
        counts = dict((id_, len(messages)) for id_, messages in raised.items())
        for id_, count in counts.items():
            self.linted_counts[id_] += count
        MessageDefManager.reset()

        # Strip messages that we're supposed to ignore. It can be tacked
        # onto the class or onto individual test methods.
        if hasattr(self, '__alinttest_ignore__'):
            # Use getattr instead of self.__alinttest_ignore__ because of
            # the mangled name.
            ignore = getattr(self, '__alinttest_ignore__')
        else:
            # Previously, we wrapped the ignorance into the test itself,
            # but then we'd forget to ignore the test during the
            # tearDown routine.
            curtest = getattr(self, self._testMethodName)
            ignore = getattr(curtest, '__alinttest_ignore__', False)

        # Something to ignore?
        if ignore:
            for key in list(counts.keys()):
                if ignore(key):
                    del counts[key]
                    # print('ignoring {} because of {}'.format(key, ignore))

        # Run this last, so we've completed the reset.
        self.assertEqual(counts, expected_counts)

    def create_instance_and_load_single_file(self, class_, filename, data):
        """
        A bit strange to have this here, but our main testing involves
        create an instance of SomeClass and loading up a sample data
        file.

        Example::

            filereader = self.create_instance_and_load_single_file(
                FileReader, 'test.conf', b'''data..''')
        """
        def opener(fn):
            assert fn == filename, (fn, filename)
            return NamedBytesIO(filename, data)
        instance = class_(opener=opener)
        instance.include(filename)
        return instance


class ALintTestResult(TextTestResult):
    def startTestRun(self):
        self.linted_counts = defaultdict(int)

    def stopTestRun(self):
        self.print_untested()

    def stopTest(self, test):
        if hasattr(test, 'linted_counts'):
            for id_, count in test.linted_counts.items():
                self.linted_counts[id_] += count
        elif test.__class__.__name__ == 'ModuleImportFailure':
            # Happens if we get module import errors during test load.
            pass
        else:
            # Otherwise, this is probably a test, and that means that
            # you didn't call our setUp()...
            raise ValueError(
                'Did you forget to call super().setUp() on {!r}?'.format(
                    test.__class__.__name__))

        super(ALintTestResult, self).stopTest(test)

    def print_untested(self):
        tested = set(self.linted_counts.keys())
        defined = MessageDefManager.types
        untested = defined - tested

        if untested:
            print(
                ('\n\nStill untested messages:\n - {}'.format(
                 '\n - '.join(sorted(list(untested))))),
                file=self.stream)


class ALintTestRunner(TextTestRunner):
    def _makeResult(self):
        return ALintTestResult(self.stream, self.descriptions, self.verbosity)


class NamedBytesIO(BytesIO):
    def __init__(self, name, data):
        super(NamedBytesIO, self).__init__(data)
        self.name = name


class _IgnoreLinted(object):
    def __init__(self, *values):
        self._equals = []
        self._startswith = []

        for value in values:
            if not isinstance(value, str):
                raise TypeError('value {!r} is not a string'.format(value))

            parts = value.split('*')
            if len(parts) == 1:
                self._equals.append(parts[0])
            elif len(parts) == 2 and not parts[1]:
                self._startswith.append(parts[0])
            else:
                raise ValueError(
                    'only an optional trailing asterisk is valid in '
                    'value {!r}'.format(value))

        self._equals = tuple(self._equals)
        self._startswith = tuple(self._startswith)
        self._repr = 'ignoreLinted({})'.format(' '.join(values))

    def __call__(self, value):
        if value in self._equals or value.startswith(self._startswith):
            return True
        return False

    def __repr__(self):
        return self._repr


def ignoreLinted(*values):
    """
    Ignore linter hints/warnings/errors as supplied. Optionally, a
    trailing asterisk (*) may be used to match more.

    Example:

        @ignoreLinted('H_DP_*', 'W_SOME_WARNING')
        class MyTest(ALintTestCase):
            ...
    """
    def decorator(test_item):
        ignorefunc = _IgnoreLinted(*values)
        test_item.__alinttest_ignore__ = ignorefunc
        return test_item
    return decorator


def main():
    orig_main(
        module=None,  # now we can put 'discover' in the command line
        testRunner=ALintTestRunner)


if __name__ == '__main__':
    main()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import string
from collections import defaultdict
from importlib import import_module

from .cls import Singleton
from .defines import ErrorDef, WarningDef
from .expression import Expr
from .function import ReadFunc, ReadFuncSlice
from .variable import Var, VarSlice, VarDynSlice
from .version import AsteriskVersion


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_FUNC_BAD_CASE(ErrorDef):
        message = 'function {func!r} does not have the proper Case {proper!r}'

    class E_FUNC_MISSING(ErrorDef):
        # BUG: this is also called for func_odbc created functions..
        message = 'function {func!r} does not exist (or func_odbc created?)'

    class E_FUNC_PARENS(ErrorDef):
        message = "missing trailing parenthesis for function call '{data}'"

    class E_FUNC_TAIL(ErrorDef):
        message = "excess tokens at the end of the function call '{data}'"

    class E_VAR_SUBSTR_ARGS(ErrorDef):
        message = "bad substring syntax of variable '{data}'"

    class E_VAR_SUBSTR_START(ErrorDef):
        message = "bad substring start value '{start}'"

    class E_VAR_SUBSTR_LENGTH(ErrorDef):
        message = "bad substring length value '{length}'"

    class W_FUNC_DYNAMIC(WarningDef):
        message = "calling functions dynamically is not good practise '{data}'"


class VarParseError(ValueError):
    pass


class FuncLoader(metaclass=Singleton):
    """
    The FuncLoader loads functions. It is called by the VarLoader when a
    function is encountered.
    """
    def __init__(self):
        self._lower_funcs = defaultdict(list)
        self._used_funcs = set()

        self.load_all()

    @property
    def used_funcs(self):
        return list(
            self._lower_funcs[i]
            for i in sorted(self._used_funcs)
            if i != 'unknown')

    @property
    def used_modules(self):
        return list(filter(
            (lambda x: x != 'unknown'),
            sorted(set(
                [self._lower_funcs[i].module
                 for i in self._used_funcs]))))

    def load_all(self):
        for mod_name in AsteriskVersion().list_func_mods():
            mod = import_module(mod_name)
            if hasattr(mod, 'register'):
                mod.register(self)

    def get(self, lower_func):
        # Fetch func named lower_func. If it doesn't exist, we alias the
        # 'UNKNOWN' func to it.
        if lower_func not in self._lower_funcs:
            if 'unknown' not in self._lower_funcs:
                raise NotImplementedError(
                    'There should be an UNKNOWN function that we can map '
                    'unknown functions to!')
            # We don't raise anything here, we do that outside, when we
            # see that the appname is not canonical.
            self._lower_funcs[lower_func] = self._lower_funcs['unknown']

        self._used_funcs.add(lower_func)
        return self._lower_funcs[lower_func]

    def register(self, func):
        lower_func = func.name.lower()
        self._lower_funcs[lower_func] = func

    def process_function(self, func_and_args, where):
        """
        Process: <function>(<args>)
        """
        # SOURCE: main/pbx.c -- ast_func_read2, func_args
        for i, char in enumerate(func_and_args):
            if char == '(':
                func = func_and_args[0:i]
                break
        assert func  # we only get here if there is a '('

        # Search backwards for the trailing ')'.
        tail = ''
        for n in range(len(func_and_args) - 1, i, -1):
            if func_and_args[n] == ')':
                args = func_and_args[(i + 1):n]
                tail = func_and_args[n + 1:]
                break
        else:
            # "Can't find trailing parenthesis for function"
            E_FUNC_PARENS(where, data=func_and_args)
            args = func_and_args[(i + 1):]

        # Do we have a tail of ":<offset>:<length>"? Process it:
        if tail:
            varname, start, length = VarLoader().split_variable_slice(
                tail, where)

            if varname:
                start = length = None
                E_FUNC_TAIL(where, data=func_and_args)
        else:
            start = length = None

        # func should be a regular string. If it isn't, you're doing
        # crazy stuff like: ${funcname}(${args})
        # I will complain.
        if isinstance(func, Var):
            # We won't be able to match any real functions. And you
            # should get a big fat warning.
            W_FUNC_DYNAMIC(where, func=func)

            if start is not None:
                return ReadFuncSlice(func, args, start=start, length=length)
            return ReadFunc(func, args)

        # Okay, so we have a string function. Look it up.
        loaded = FuncLoader().get(func.lower())

        # Check function availability.
        if loaded.name != func:
            if loaded.name == 'UNKNOWN':
                E_FUNC_MISSING(where, func=func)
            else:
                E_FUNC_BAD_CASE(where, func=func, proper=loaded.name)

        # TODO: do more stuff with this..?

        if start is not None:
            return ReadFuncSlice(func, args, start=start, length=length)
        return ReadFunc(func, args)


class VarLoader(metaclass=Singleton):
    """
    The VarLoader loads variables, functions and expressions. It's
    contained here with the variables only, because the functions are
    a special form of variables in asterisk.
    """
    def __init__(self):
        self._variables = defaultdict(list)

    def parse_variables(self, data, where):
        """
        Parse variables will loop over the data and extract all
        variables. This function mimic the behaviour of Asterisk
        variable substitution, except we don't do the substitution just
        yet. (Because we don't know the values of the variables.)
        """
        # SOURCE: main/pbx.c -- ast_str_substitute_variables_full
        # SOURCE: main/pbx.c -- pbx_substitute_variables_full
        ret = []
        beginpos = searchpos = 0
        while True:
            try:
                pos = data.index('$', searchpos)
                next_ = data[pos + 1]
            except (IndexError, ValueError):
                ret.append(data[beginpos:])
                break
            else:
                if next_ in '{[':
                    ret.append(data[beginpos:pos])
                    if next_ == '{':
                        endpos, more_substitutions = self._find_brackets_end(
                            data, pos + 2, '{', '}')
                    elif next_ == '[':
                        endpos, more_substitutions = self._find_brackets_end(
                            data, pos + 2, '[', ']')

                    inner_data = data[(pos + 2):(endpos - 1)]
                    if more_substitutions:
                        inner_data = self.parse_variables(
                            inner_data, where)

                    if next_ == '{':
                        inner_data = self._process_variable(
                            inner_data, where)
                    elif next_ == '[':
                        inner_data = self._process_expression(
                            inner_data, where)

                    ret.append(inner_data)
                    data = data[endpos:]
                    beginpos = searchpos = 0
                else:
                    searchpos = pos + 1

        return Var.join(ret)

    def _find_brackets_end(self, data, pos, beginbracket, endbracket):
        more_substitutions = False
        brackets = 1
        while brackets:
            try:
                ch = data[pos]
            except IndexError:
                raise VarParseError(
                    'Error in extension logic (missing {!r})'.format(
                        endbracket))
            if ch == '$':
                if data[(pos + 1):(pos + 2)] in '{[':
                    more_substitutions = True
            elif ch == beginbracket:
                brackets += 1
            elif ch == endbracket:
                brackets -= 1
            pos += 1
        return pos, more_substitutions

    def _process_variable(self, data, where):
        """
        Process: <varname>

        (Or sliced variable, or expression, or function.)
        """
        if '(' in data:
            # Create function and return.
            return FuncLoader().process_function(data, where)

        elif ':' in data:
            # Create sliced var and return.
            return self._process_variable_slice(data, where)

        else:
            # Count Var usage.
            if not isinstance(data, Var):
                self.count_var(data, where)

            # On to return something sensible.
            return Var(data)

    def _process_variable_slice(self, data, where):
        """
        Process: <varname>:<start>[:<length>]
        """
        varname, start, length = self.split_variable_slice(data, where)

        # Count Var usage.
        if not isinstance(varname, Var):
            self.count_var(varname, where)

        # On to return something sensible.
        if start is None:
            return Var(varname)

        # Very dynamic, this..
        if isinstance(start, Var) or isinstance(length, Var):
            return VarDynSlice(varname, start=start, length=length)

        return VarSlice(varname, start=start, length=length)

    def _process_expression(self, data, where):
        """
        Process: <expression>
        """
        # We should parse the expression and supply errors/warnings
        # here, before returning the expression.
        return Expr(data)

    def count_var(self, varname, where):
        assert isinstance(varname, str), varname
        assert all((i in string.ascii_letters or
                    i in string.digits or
                    i == '_')
                   for i in varname), varname
        self._variables[varname].append(where)

    @staticmethod
    def split_variable_slice(data, where):
        """
        Split up: <varname>:<start>[:<length>] and return the three
        parts.
        """
        parts = data.split(':')
        if len(parts) == 2:
            varname, start, length = parts[0], parts[1], None
        elif len(parts) == 3:
            varname, start, length = parts[0], parts[1], parts[2]
        else:
            varname = parts[0]
            start = length = None
            E_VAR_SUBSTR_ARGS(where, data=data)

        if start is not None:
            if isinstance(start, Var):
                # We cannot determine anything more from this.
                pass
            elif (start and (
                    start.isdigit() or
                    (start[0] == '-' and start[1:].isdigit()))):
                start = int(start)
            else:
                start = length = None
                E_VAR_SUBSTR_START(where, start=start)

        if length is not None:
            if isinstance(length, Var):
                # We cannot determine anything more from this.
                pass
            elif (length and (
                    length.isdigit() or
                    (length[0] == '-' and length[1:].isdigit()))):
                length = int(length)
                # If we use an offset from the end, then it makes no
                # sense to have a length that's as large or larger.
                if length == 0 or (start < 0 and -length <= start):
                    length = None
                    E_VAR_SUBSTR_LENGTH(where, length=length)
            else:
                E_VAR_SUBSTR_LENGTH(where, length=length)
                start = length = None

        # If start is 0 and there is no length, it makes no sense.
        if start == 0 and not length:
            start = None
            E_VAR_SUBSTR_START(where, start=start)

        return varname, start, length



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .defines import ErrorDef, WarningDef
from .file import W_WSH_BOL


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class I_NOTIMPL_TEMPLATES(ErrorDef):
        message = 'asterisklint does not implement template use yet'

    class W_WSH_OBJSET(WarningDef):
        message = 'expected " => " horizontal whitespace around arrow operator'

    class W_WSH_VARSET(WarningDef):
        message = 'expected no horizontal whitespace around equals operator'


class EmptyLine(object):
    def __init__(self, comment, where):
        self.comment = comment
        self.where = where


class Context(object):
    @classmethod
    def from_context(cls, context, **kwargs):
        """
        Use this on subclasses of Context.
        """
        assert not context._varsets
        return cls(name=context.name, templates=context._templates,
                   comment=context.comment, bolspace='',
                   where=context.where, **kwargs)

    def __init__(self, name, templates='', comment=False, bolspace='',
                 where=None):
        if bolspace:
            W_WSH_BOL(where)

        self.name = name
        self.comment = comment
        self.where = where
        self._templates = templates
        self._varsets = []

        if templates:
            I_NOTIMPL_TEMPLATES(where)

    def update(self, othercontext):
        assert self.name == othercontext.name
        assert not othercontext.comment
        self.where = othercontext.where  # or not?
        assert not othercontext._templates
        assert not othercontext._varsets

    def add(self, varset):
        self._varsets.append(varset)

    def __bool__(self):
        # Must(!) define this, now that we use __len__.
        return True

    def __len__(self):
        return len(self._varsets)

    def __getitem__(self, key):
        assert isinstance(key, int)
        return self._varsets[key]

    def __repr__(self):
        return '[{}]({}) => ({} elements)'.format(
            self.name, self._templates, len(self._varsets))


class Varset(object):
    def __init__(self, variable, value, separator, comment, where):
        clean_separator = separator.strip()
        if clean_separator == '=>':
            if separator != ' => ':
                W_WSH_OBJSET(where)
            self.arrow = True
        elif clean_separator == '=':
            if separator != '=':
                W_WSH_VARSET(where)
            self.arrow = False

        if variable.startswith(tuple(' \t')):
            # QUICK HACK: only allow leading WS for 'same'..
            variable = variable.lstrip(' \t')
            if variable != 'same':
                W_WSH_BOL(where)

        self.variable = variable
        self.value = value
        self.comment = comment
        self.where = where



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from .cls import Singleton


class AsteriskVersion(metaclass=Singleton):
    """
    Store the used Asterisk version globally. If you don't initialize
    this before anyone requests anything, you get the default.

    Example::

        from asterisklint.version import AsteriskVersion

        AsteriskVersion('v11')  # set version 11 throughout the run
    """
    DEFAULT = 'v11'

    def __init__(self, version=None):
        self.version = version or self.DEFAULT

    def reinit(self, version=None):
        if version and self.version != version:
            raise RuntimeError(
                'Attempt to re-set Asterisk version from {} to {}'.format(
                    self.version, version))

    def list_app_mods(self):
        """
        Return a list app names in absolute import format. Takes
        internal version into account.
        """
        return self._list_mods('app')

    def list_func_mods(self):
        """
        Return a list function names in absolute import format. Takes
        internal version into account.
        """
        return self._list_mods('func')

    def _get_path(self, submod):
        return os.path.join(os.path.dirname(__file__), submod, self.version)

    def _list_mods(self, submod):
        appsdir = self._get_path(submod)
        appmods = [i[0:-3] for i in os.listdir(appsdir) if i.endswith('.py')]

        modfmt = 'asterisklint.{}.{}.{{}}'.format(submod, self.version)
        return [modfmt.format(appmod) for appmod in appmods]



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import AppBase, IfStyleApp


class _Builtin(object):
    @property
    def module(self):
        return '<builtin>'


class BuiltinAppBase(_Builtin, AppBase):
    pass


class Answer(BuiltinAppBase):
    pass


class Background(BuiltinAppBase):
    # NOTE: It is called "BackGround" in Asterisk, but that makes my
    # eyes sore.
    pass


class Busy(BuiltinAppBase):
    pass


class Congestion(BuiltinAppBase):
    pass


class Goto(BuiltinAppBase):
    pass


class GotoIf(_Builtin, IfStyleApp):
    pass


class GotoIfTime(_Builtin, IfStyleApp):
    pass


class ExecIfTime(_Builtin, IfStyleApp):
    pass


class Hangup(BuiltinAppBase):
    pass


class NoOp(BuiltinAppBase):
    pass


class Proceeding(BuiltinAppBase):
    pass


class Progress(BuiltinAppBase):
    pass


class ResetCDR(BuiltinAppBase):
    pass


class Ringing(BuiltinAppBase):
    pass


class SayNumber(BuiltinAppBase):
    pass


class SayDigits(BuiltinAppBase):
    pass


class SayAlpha(BuiltinAppBase):
    pass


class SayPhonetic(BuiltinAppBase):
    pass


class SetAMAFlags(BuiltinAppBase):
    pass


class SetGlobalVar(BuiltinAppBase):
    pass


class Set(BuiltinAppBase):
    pass


class ImportVar(BuiltinAppBase):
    pass


class Wait(BuiltinAppBase):
    pass


class WaitExten(BuiltinAppBase):
    pass


def register(app_loader):
    for app in (
            Answer, Background, Busy,
            Congestion, Goto, GotoIf,
            GotoIfTime, ExecIfTime,
            Hangup, NoOp, Proceeding,
            Progress, ResetCDR, Ringing,
            SayNumber, SayDigits, SayAlpha,
            SayPhonetic, SetAMAFlags,
            SetGlobalVar, Set,
            ImportVar, Wait,
            WaitExten):
        app_loader.register(app())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import os

from . import FileFuncOdbcParser


class UniqueStore(argparse.Action):
    "Make sure an --argument is only used at most once."
    def __call__(self, parser, namespace, values, option_string):
        if getattr(namespace, self.dest, self.default) is not None:
            parser.error("{} appears several times".format(option_string))
        setattr(namespace, self.dest, values)


def load_func_odbc_functions(func_odbc_arg, dialplan_arg):
    if func_odbc_arg == '':
        # Explicit disable.
        return

    if func_odbc_arg is None:
        # Auto-detect?
        func_odbc_arg = os.path.join(
            os.path.dirname(dialplan_arg), 'func_odbc.conf')
        if not os.path.exists(func_odbc_arg):
            # If there is no such file, don't worry about it.
            return

    # Load it up.
    # TODO: do we want to silence func_odbc errors/warnings here?
    parser = FileFuncOdbcParser()
    parser.include(func_odbc_arg)
    for func_odbc_context in parser:
        pass



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .configobj import EmptyLine, Context, Varset
from .defines import HintDef
from .file import W_WSV_BOF, W_WSV_EOF


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class H_WSV_CTX_BETWEEN(HintDef):
        # We skip this if the context contains only 0..2 varsets.
        message = 'expected one or two lines between non-tiny contexts'

    class H_WSV_VARSET_BETWEEN(HintDef):
        message = 'expected zero or one lines between varsets'


class VswContext(object):
    """
    We need a separate context object to keep state per file. Because of
    how the parser works, we won't get any notifications about file
    switches, other than that the "where" changes. This VswContext and
    the VswContextStack keep track of the changes and of the
    file-specific state.
    """
    def __init__(self, where):
        self.filename = where.filename
        self.bof = True
        self.blanks = []
        self.last_context_varsets = 0

    def is_where(self, where):
        return self.filename == where.filename

    def handle(self, element):
        if isinstance(element, EmptyLine):
            self.blanks.append(element)
        else:
            for delayed in self.handle_nonblank(element):
                yield delayed

    def handle_nonblank(self, element):
        if isinstance(element, Context):
            if self.last_context_varsets <= 2:
                min_blanks = 0
            else:
                min_blanks = 1
            max_blanks = 2
            message = H_WSV_CTX_BETWEEN
            self.last_context_varsets = 0
        elif isinstance(element, Varset):
            min_blanks, max_blanks = 0, 1
            message = H_WSV_VARSET_BETWEEN
            self.last_context_varsets += 1
        else:
            raise NotImplementedError('unknown element {!r}'.format(
                element))

        for blank in self.flush_blanks(element, message, min_blanks,
                                       max_blanks):
            yield blank

        self.bof = False  # after flush_blanks..
        yield element

    def flush_blanks(self, element, message, min_blanks, max_blanks):
        # Check minvalues first.
        if len(self.blanks) < min_blanks:
            message(element.where)
            for blank in self.blanks:
                yield blank
            self.blanks = []
            return

        # Nothing to do?
        if not self.blanks:
            return

        # Group the blanks by comment/non-comment. From the
        # non-comments we only want to see max_blanks in a row.
        grouped = self.group_blanks(self.blanks)
        if self.bof:
            if not grouped[0][0]:
                W_WSV_BOF(grouped[0][1][0].where)
                has_comment, grouped_blanks = grouped.pop(0)
                for blank in grouped_blanks:
                    yield blank
        for i, (has_comment, grouped_blanks) in enumerate(grouped):
            if (not has_comment and
                    len(grouped_blanks) > max_blanks):
                message(grouped_blanks[-1].where)
            for blank in grouped_blanks:
                yield blank
        self.blanks = []

    def handle_eof(self, report_eof_blanks=True):
        if not self.blanks:
            return

        # Group any leftover blanks by comment/non-comment.
        message = H_WSV_CTX_BETWEEN
        max_blanks = 2
        grouped = self.group_blanks(self.blanks)

        # Save the tail for the EOF warning.
        if not grouped[-1][0]:
            eof_blanks = grouped.pop()[1]
        else:
            eof_blanks = []

        # Process the rest.
        for has_comment, grouped_blanks in grouped:
            if (not has_comment and
                    len(grouped_blanks) > max_blanks):
                message(grouped_blanks[-1].where)
            # Yield it all.
            for blank in grouped_blanks:
                yield blank

        # Check tail whitespace and report on that.
        if eof_blanks and report_eof_blanks:
            for blank in eof_blanks:
                yield blank
            W_WSV_EOF(self.blanks[-1].where)

    def group_blanks(self, blanks):
        ret = []
        for blank in blanks:
            if ret and ret[-1][0] == bool(blank.comment):
                ret[-1][1].append(blank)
            else:
                ret.append((bool(blank.comment), [blank]))
        return ret


class VswContextStack(list):
    def handle(self, element):
        if not self.is_on_stack(element.where):
            self.append(VswContext(element.where))
        else:
            while not self[-1].is_where(element.where):
                locontext = self.pop()
                for delayed in locontext.handle_eof():
                    yield delayed
            assert self, 'eof of last file will not happen here'

        for delayed in self[-1].handle(element):
            yield delayed

    def is_on_stack(self, where):
        return any(i.is_where(where) for i in reversed(self))

    def handle_eof(self):
        first = True
        while self:
            locontext = self.pop()
            for delayed in locontext.handle_eof(report_eof_blanks=first):
                yield delayed
            first = False


class VerticalSpaceWarner(object):
    """
    Warns about irregular vertical whitespace (WSV).

    Contexts get one or two leading white lines. Varsets get at most one
    leading white line.
    """
    def __iter__(self):
        # We use a VswContext object to keep state per "included" file.
        # Every time a file include changes, we push/pop the context
        # stack.
        stack = VswContextStack()
        for element in super().__iter__():
            for delayed in stack.handle(element):
                yield delayed
        for delayed in stack.handle_eof():
            yield delayed



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show which modules, apps and functions are used by the dialplan. Takes
'extensions.conf' as argument.
"""
import argparse
from collections import defaultdict

from asterisklint import FileDialplanParser
from asterisklint.application import E_APP_MISSING, AppLoader
from asterisklint.defines import MessageDefManager
from asterisklint.mainutil import UniqueStore, load_func_odbc_functions
from asterisklint.varfun import E_FUNC_MISSING, FuncLoader


class Aggregator(object):
    """
    Collector of all E_APP_MISSING, E_FUNC_MISSING errors. We want to
    know about these things.
    """
    def __init__(self):
        E_APP_MISSING.add_callback(self.on_e_app_missing)
        E_FUNC_MISSING.add_callback(self.on_e_func_missing)
        self.unknown_apps = set()
        self.unknown_funcs = set()

    def on_e_app_missing(self, message):
        self.unknown_apps.add(message.fmtkwargs['app'])

    def on_e_func_missing(self, message):
        self.unknown_funcs.add(message.fmtkwargs['func'])


def main(args, envs):
    parser = argparse.ArgumentParser(
        description=(
            "Show which modules, apps and functions are used by the dialplan. "
            "Useful when you use autoload=no in your modules.conf. Beware "
            "that you do need more modules than just these listed."))
    parser.add_argument(
        'dialplan', metavar='EXTENSIONS_CONF',
        help="path to extensions.conf")
    parser.add_argument(
        '--func-odbc', metavar='FUNC_ODBC_CONF', action=UniqueStore,
        help="path to func_odbc.conf, will be read automatically if found "
             "in same the same dir as extensions.conf; set empty to disable")
    args = parser.parse_args(args)

    # No messages to stderr, but do collect the E_APP_MISSING and
    # E_FUNC_MISSING errors.
    aggregator = Aggregator()
    MessageDefManager.muted = True

    # Load func_odbc functions if requested.
    load_func_odbc_functions(args.func_odbc, args.dialplan)

    parser = FileDialplanParser()
    parser.include(args.dialplan)
    dialplan = next(iter(parser))
    del dialplan

    apploader = AppLoader()
    funcloader = FuncLoader()
    all_modules = set()

    for what, used_items, used_modules in (
            ('Application', apploader.used_apps, apploader.used_modules),
            ('Function', funcloader.used_funcs, funcloader.used_modules)):
        all_modules.update(used_modules)

        used_items_per_module = defaultdict(list)
        for item in used_items:
            used_items_per_module[item.module].append(item)

        print('; {} providing modules used:'.format(what))
        for module in used_modules:
            items = used_items_per_module[module][:]
            item_lines = ['']
            while items:
                next_ = items.pop(0)
                if item_lines[-1]:
                    item_lines[-1] += ', '
                formatted = '{}()'.format(next_.name)
                if len(item_lines[-1]) + len(formatted) > 52:
                    item_lines.append(formatted)
                else:
                    item_lines[-1] += formatted

            # Output.
            print(';   {:20s}  {}'.format(
                module, item_lines[0].strip()))
            for item_line in item_lines[1:]:
                print(';   {:20s}  {}'.format(
                    '', item_line.strip()))
        print(';')

    print('; modules.conf')
    for module in sorted(all_modules):
        if module != '<builtin>':
            print('load => {}.so'.format(module))
    print()
    if aggregator.unknown_apps:
        print('; WARNING: The following unknown applications were seen:')
        print(';   {}'.format(', '.join(sorted(aggregator.unknown_apps))))
        print(';')
    if aggregator.unknown_funcs:
        print('; WARNING: The following unknown functions were seen:')
        print(';   {}'.format(', '.join(sorted(aggregator.unknown_funcs))))
        print(';')

    return int(bool(aggregator.unknown_apps) or bool(aggregator.unknown_funcs))



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Singleton(type):
    """
    Create singleton classes. Usage::

        class MyClass(metaclass=Singleton):
            ...

        MyClass().do_stuff()

    If your class implements the reinit() method, it will be called with
    the same arguments as to __init__ in case it was already created.
    There you may evaluate if the args/kwargs are allowed and/or need
    updating::

        def reinit(self, optional_config=None):
            if optional_config and self.config != optional_config:
                raise ProgrammingError("You cannot switch config!")
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            ret = cls._instances[cls] = super().__call__(*args, **kwargs)
        else:
            ret = cls._instances[cls]
            if hasattr(ret, 'reinit'):
                ret.reinit(*args, **kwargs)

        return ret



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from importlib import import_module

from .cls import Singleton
from .defines import ErrorDef, WarningDef
from .varfun import VarLoader, VarParseError
from .version import AsteriskVersion


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_APP_WSH(ErrorDef):
        message = 'whitespace before app {app!r} will result in unknown app'

    class E_APP_MISSING(ErrorDef):
        message = 'app {app!r} does not exist, dialplan will halt here!'

    class E_APP_PARSE_ERROR(ErrorDef):
        message = 'app {app!r} arguments {args!r} raise a parse error'

    class W_APP_BALANCE(WarningDef):
        message = ('app data {data!r} looks like unbalanced'
                   'parentheses/quotes/curlies')

    class W_APP_BAD_CASE(WarningDef):
        message = 'app {app!r} does not have the proper Case {proper!r}'

    class W_APP_NEED_PARENS(WarningDef):
        message = 'app {app!r} should have parentheses (only NoOp is exempt)'

    class W_APP_WSH(ErrorDef):
        message = 'unexpected whitespace after app {app!r}'


class AppLoader(metaclass=Singleton):
    def __init__(self):
        self._lower_apps = {}
        self._used_apps = set()

        self.load_all()

    @property
    def used_apps(self):
        return list(
            self._lower_apps[i]
            for i in sorted(self._used_apps)
            if i != 'unknown')

    @property
    def used_modules(self):
        return list(filter(
            (lambda x: x != 'unknown'),
            sorted(set(
                [self._lower_apps[i].module
                 for i in self._used_apps]))))

    def load_all(self):
        for mod_name in AsteriskVersion().list_app_mods():
            mod = import_module(mod_name)
            if hasattr(mod, 'register'):
                mod.register(self)

    def get(self, lower_app):
        # Fetch app named lower_app. If it doesn't exist, we alias the
        # 'Unknown' app to it.
        if lower_app not in self._lower_apps:
            if 'unknown' not in self._lower_apps:
                raise NotImplementedError(
                    'There should be an Unknown application that we can map '
                    'unknown applications to!')
            # We don't raise anything here, we do that outside, when we
            # see that the appname is not canonical.
            self._lower_apps[lower_app] = self._lower_apps['unknown']

        self._used_apps.add(lower_app)
        return self._lower_apps[lower_app]

    def register(self, app):
        lower_app = app.name.lower()
        self._lower_apps[lower_app] = app


class App(object):
    # App heeft weer z'n eigen parsers en subparsers. Hiermee moeten we ook
    # op kunnen zoeken welke modules er nodig zijn (w00t). Deze komt ook als
    # eerste in aanmerking voor 1.4 vs. 1.8 differences (ExecIf..., vs
    # ExecIf...?)
    # Ook dumpen van alle variabelen (in dat geval moet de subparser aangeven
    # dat een Var geset wordt (Set() en MSet() doen dat bijv.) en ARRAY() en
    # HASH().
    # Parser en subparsers implementeren volgens included modules?
    # Klinkt wat overkill? Maar maakt het wel mooi extensible. De ExecIf()
    # kan dan de args weer splitten en teruggeven...
    #
    # Per subparser kunnen we de versienummers opgeven:
    # SPRINTF() voor < 1.2 => None
    # SPRINTF() voor >= 1.2 < 1.4 => behaviour X
    # SPRINTF() voor >= 1.4 => behaviour Y
    #
    # Hoe pakken we dan app_compat settings voor Set? Hm.
    def __init__(self, app, where):
        self.raw = app
        self.where = where

        # Attempt to parse the app + data.
        self.parse()

    def parse(self):
        """
        Parse self.raw into an App and data. Recursively replaces the
        app data as far as possible.

        It tries to find a specific handler for the application, but
        will fall back to a default handler (parse_simple) if it cannot
        be found.
        """
        # Fetch the Application. If we don't know specifically, we can
        # call the parse_simple on it.
        if not self.split_app_data():
            return

        # Find the handler from the registered handlers. If there is no
        # custom handler, we may already raise a message here.
        app = AppLoader().get(self.app_lower)

        # Pass the data through a handler -- which also handles
        # functions -- first:
        self.data = self.parse_inner(self.data)

        # Check app availability.
        if app.name != self.app:
            if app.name == 'Unknown':
                E_APP_MISSING(self.where, app=self.app)
            else:
                W_APP_BAD_CASE(self.where, app=self.app, proper=app.name)

        # Run data through app.
        app(self.data, where=self.where)

    def parse_inner(self, data):
        """
        Asterisk calls pbx_substitute_variables_helper_full on the
        entire app "data" line. So do we.
        """
        # SOURCE: main/pbx.c: pbx_extension_helper, calls
        # SOURCE: main/pbx.c: pbx_parse_variables_helper_full.
        if '${' in data or '$[' in data:
            try:
                data = VarLoader().parse_variables(data, self.where)
#            except NotImplementedError:  # FIXME: remove this
#                pass
            except VarParseError:
                E_APP_PARSE_ERROR(self.where, app=self.app, args=data)
        return data

    def split_app_data(self):
        """
        Splits self.raw into self.app, self.app_lower, self.data.
        """
        try:
            app, data = self.raw.split('(', 1)
        except ValueError:
            # We allow NoOp without parentheses. The others need parens.
            app = self.raw
            if self.raw.lower() != 'noop':
                W_APP_NEED_PARENS(self.where, app=app)
            data = '()'
        else:
            data = '(' + data

        # SOURCE: pbx/pbx_config.c: pbx_load_config()
        if data.startswith('(') and data.endswith(')'):
            data = data[1:-1]

        # Set it and try to find a handler for it.
        self.app = app
        self.app_lower = app.lower()
        self.data = data

        # Leading whitespace is frowned upon but allowed. Trailing
        # whitespace won't work:
        if self.app.rstrip() != self.app:
            E_APP_WSH(self.where, app=self.app)
            return False
        if self.app.lstrip() != self.app:
            W_APP_WSH(self.where, app=self.app)
            self.app = self.app.lstrip()
            self.app_lower = self.app.lower()
        # Quick check that the app doesn't exist.
        if not self.app:
            E_APP_MISSING(self.where, app='(none)')
            return False

        return True

    @property
    def name(self):
        return self.app



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import importlib
import os
import re
import sys

from . import alintver


COMMAND_RE = re.compile(r'^[a-z0-9_][a-z0-9_-]*$')


class NoSuchCommand(ValueError):
    pass


def wrap(text, skip, maxlen):
    lines = []
    words = text.split()
    maxlen -= (skip + 1)  # add one for the space
    for word in words:
        if not lines:
            lines.append(word)
        elif len(lines[-1]) + len(word) >= maxlen:
            lines.append(word)
        else:
            lines[-1] += ' ' + word
    return '\n{}'.format(' ' * skip).join(lines)


def list_commands(args, envs):
    """
    List the available commands by browsing through your pythonpath and
    listing the commands inside 'asterisklint.commands'.
    """
    commands_used = set()
    commands = []
    for path in sys.path:
        try:
            files = os.listdir(os.path.join(path, 'asterisklint', 'commands'))
        except OSError:
            pass
        else:
            for file_ in files:
                if file_.endswith('.py') and COMMAND_RE.match(file_[0:-3]):
                    command_name = file_[0:-3]
                    if command_name not in commands_used:
                        try:
                            command = load_command(command_name)
                        except NoSuchCommand as e:
                            command_desc = '(error: {})'.format(e)
                        else:
                            try:
                                command_desc = command.__doc__.strip()
                            except AttributeError:
                                command_desc = '(help text missing)'

                        commands_used.add(command_name)
                        commands.append((path, command_name, command_desc))
    commands.sort()

    fmt = '  {name:21} {desc}'
    print('builtin:')
    print(fmt.format(name='ls', desc='List available commands.'))
    last_path = None
    for path, name, desc in commands:
        if last_path != path:
            print()
            print('{}:'.format(path))
            last_path = path
        print(fmt.format(name=name, desc=wrap(desc, 24, 80)))
    print()
    print('Place custom commands in ~/.asterisklint/asterisklint/commands.')


def load_command(command):
    """
    Load ``command`` from the asterisklint.command namespace and return it.

    The command could be placed anywhere in your pythonpath, as long as
    it's inside the asterisklint.commands package.
    """
    if not COMMAND_RE.match(command):
        raise NoSuchCommand('Invalid tokens in command {!r}'.format(command))

    try:
        command_module = importlib.import_module(
            'asterisklint.commands.{}'.format(command))
    except ImportError as exception:
        raise NoSuchCommand('Could not load {!r}: {}'.format(
            command, exception))

    try:
        command_module.main
    except AttributeError:
        raise NoSuchCommand('Command {!r} has no entrypoint'.format(command))

    return command_module


def main(args, envs):
    # Insert ~/.asterisklint into path if it exists. This path construct
    # will also allow you to override any part of the application; like
    # add/remove individual Asterisk apps or functions.
    userdir = '{}/.asterisklint'.format(envs.get('HOME', '/dev/null'))
    if os.path.exists(userdir) and userdir not in sys.path:
        sys.path.insert(0, userdir)

    # Fetch first argument: the command.
    non_options = [i for i in args if not i.startswith('-')]
    if non_options:
        command = non_options[0]
        args.remove(command)

        # Builtin commands.
        if command == 'ls':
            list_commands(args, envs)

        # Dynamically loaded commands.
        else:
            try:
                command_module = load_command(command)
            except NoSuchCommand as e:
                print(e, file=sys.stderr)
                return 1

            # Update argv[0] for the command argparse help.
            sys.argv[0] = '{} {}'.format(sys.argv[0], command)
            # Run command main.
            return command_module.main(args, envs)

    # If there was no command, pass it along to the arg parser.
    else:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description='Check Asterisk PBX configuration syntax.')
        parser.add_argument(
            '-V', '--version', action='version',
            version="asterisklint {0.version_str}\n\n{0.license_str}".format(
                alintver))
        parser.add_argument(
            'command', metavar='COMMAND',
            help=("the command to execute, 'ls' lists the available "
                  "commands"))

        args = parser.parse_args(args)  # will fail/exit, because .. no command



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..vall.builtin import register as builtin_register
from ..vall.unknown import register as unknown_register

from ..vall.app_confbridge import register as app_confbridge_register
from ..vall.app_queue import register as app_queue_register
from ..vall.chan_sip import register as chan_sip_register
from ..vall.func_callerid import register as func_callerid_register
from ..vall.func_cdr import register as func_cdr_register
from ..vall.func_channel import register as func_channel_register
from ..vall.func_curl import register as func_curl_register
from ..vall.func_cut import register as func_cut_register
from ..vall.func_enum import register as func_enum_register
from ..vall.func_env import register as func_env_register
from ..vall.func_global import register as func_global_register
from ..vall.func_groupcount import register as func_groupcount_register
from ..vall.func_logic import register as func_logic_register
from ..vall.func_math import register as func_math_register
from ..vall.func_rand import register as func_rand_register
from ..vall.func_shell import register as func_shell_register
from ..vall.func_strings import register as func_strings_register
from ..vall.func_uri import register as func_uri_register
from ..vall.res_fax import register as res_fax_register


def register(func_loader):
    for regfunc in (
            builtin_register,
            unknown_register,
            app_confbridge_register,
            app_queue_register,
            chan_sip_register,
            func_callerid_register,
            func_cdr_register,
            func_channel_register,
            func_curl_register,
            func_cut_register,
            func_enum_register,
            func_env_register,
            func_global_register,
            func_groupcount_register,
            func_logic_register,
            func_math_register,
            func_rand_register,
            func_shell_register,
            func_strings_register,
            func_uri_register,
            res_fax_register):
        regfunc(func_loader)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Do sanity checks on func_odbc.conf. Takes 'func_odbc.conf' as argument.
Suppress error classes using ALINT_IGNORE.
"""
import argparse

from asterisklint import FileFuncOdbcParser
from asterisklint.defines import MessageDefManager


def main(args, envs):
    parser = argparse.ArgumentParser(
        description=(
            'Do sanity checks on dialplan. Suppress comma separated '
            'error classes through the ALINT_IGNORE environment variable. '
            'Returns 1 if any issue was reported.'))
    parser.add_argument(
        'func_odbc', metavar='FUNC_ODBC_CONF',
        help="path to func_odbc.conf")
    args = parser.parse_args(args)

    parser = FileFuncOdbcParser()
    parser.include(args.func_odbc)
    for func_odbc_context in parser:
        pass

    # MessageDefManager.raised is a dict of messages ordered by message
    # type. All message types share the same muted flag, so we need only
    # examine the first.
    if any(not i[0].muted for i in MessageDefManager.raised.values()):
        return 1



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from functools import total_ordering
from re import compile as re_compile

from .defines import HintDef


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class H_PAT_NON_CANONICAL(HintDef):
        message = 'pattern {pat!r} is not in the canonical form {expected!r}'

    # TODO: two classes of errors?
    # - in-pattern errors
    # - in-dialplan errors (inconsistent used of same-valued pattern, or
    #   could we catch this by using a single canonical form only?)


# Pattern info:
#
# X Z N   match ranges 0-9, 1-9, 2-9 respectively.
# [       denotes the start of a set of character. Everything inside
#         is considered literally. We can have ranges a-d and individual
#         characters. A '[' and '-' can be considered literally if they
#         are just before ']'.
#         (currently there is no way to specify ']' in a range, nor \ is
#         considered specially.)
#
# General problems:
#
# - trip on lowercase (tackled by canonical form)
# - trip on inconsistent use of pattern styles (combining [0-9] with X)
#   (tackled by canonical form)
# - trip on backslashes and spaces in patterns
# - trip on unescaped undelimited XZN between other letters
#   (_snake-! == _s[0-9]ake-!, should be _s[n]ake-!)
# - trip on empty range, bracket in range, dash in range, backslash..
# - ...
#
# Pattern order:
#
# Patterns are ordered from left to right, by matchcharacter, where a
# matchcharacter can be one of these.
# - 0x001xx     one character, character set starting with xx
# - 0x0yyxx     yy characters, character set starting with xx
# - 0x18000     '.' (one or more of anything)
# - 0x28000     '!' (zero or more of anything)
# - 0x30000     NUL (end of string)
# - 0x40000     error in set.
#
# However(!), when sorting, the non-pattern comes before a pattern
# (see ext_cmp()).


@total_ordering  # now we only need 'eq' and 'lt'
class Pattern(object):
    NOT_A_PATTERN = 0
    IS_A_PATTERN = 1

    RE_NO_DASH = re_compile(r'(\[.[^]]*\]|[^][-]+)')

    # Pattern heeft equality tests zodat "s-zap" == "s-[1-9]ap", maar emit
    # wel een warning als je hier iets anders neerzet! (Zelfde verhaal
    # met hoofdletters vs. kleine letters.)
    def __init__(self, pattern, where):
        self.raw = pattern
        self.where = where

        # Values should hold an array of unsigneds:
        # (u32, u32, u32, ...)
        #
        # There the first integer defines the pattern type (value or
        # pattern) and the second value is a single integral value as
        # mentioned in the pattern value-list above.
        #
        # 100     = (0, 0x31, 0x030, 0x30, 0x30000)
        # _20     = (1, 0x32, 0x031, 0x30000)
        # _2[0-2] = (1, 0x32, 0x330, 0x30000)
        #
        # Unless the pattern has gaps, in which case we store it as a
        # tuple instead.
        #
        # _2[013] = (1, (0x32,), (0x330, b'013'), (0x4000,))
        #
        # This would require us to marshal patterns to the same type
        # before comparing the values. And that complicates things.
        # Let's live with the space complexity for now and cast all
        # to a list of tuples.
        #
        # We could consider investing in making the Pattern a Fly-Weight
        # or Singleton, so it doesn't consume more memory than it has
        # to. (Check sys.getsizeof() for actual space usage.)
        #
        self.values = self.parse(pattern)

    @classmethod
    def parse(cls, raw):
        "Takes str, returns tuple."
        if not raw:
            return tuple(-1, (0x40000,))  # bad pattern

        if raw[0] == '_':
            return tuple([Pattern.IS_A_PATTERN] + cls.parse_pattern(raw[1:]))

        return tuple(
            [Pattern.NOT_A_PATTERN] +
            [(0x100 + i,) for i in bytes(raw, 'utf-8') if i != 0x2d])

    @classmethod
    def parse_pattern(cls, raw):
        "Takes str, returns list."""
        raw = list(bytes(raw, 'utf-8'))
        ret = []
        while raw:
            num = raw.pop(0)
            if num == 0x2d:             # '-'
                pass
            elif num in (0x58, 0x78):   # 'X'/'x'
                ret.append((0xa30,))    # (10 * 0x100) + ord('0')
            elif num in (0x5a, 0x7a):   # 'Z'/'z'
                ret.append((0x931,))    # (9 * 0x100) + ord('1')
            elif num in (0x4e, 0x6e):   # 'N'/'n'
                ret.append((0x832,))    # (8 * 0x100) + ord('2')
            elif num == 0x2e:           # '.'
                ret.append((0x18000,))
            elif num == 0x21:           # '!'
                ret.append((0x28000,))
            elif num == 0x5b:           # '['
                try:
                    range_end = raw.index(0x5d)  # ']'
                except IndexError:
                    # TODO: raise error!
                    ret.append((0x40000,))
                    return ret
                ret.append(cls.parse_range_list(raw[0:range_end]))
                raw = raw[(range_end + 1):]  # drop ']'
            else:
                # TODO: warn on closing bracket or other non-standard
                # characters
                ret.append((0x100 + num,))

        ret.append((0x30000,))
        return ret

    @classmethod
    def parse_range_list(cls, rawlist):
        "Takes list, returns number."""
        # TODO: check empty list?
        # TODO: complain about backslashes?
        # TODO: warn on [-0-9] (prefer [0-9-])
        usedlist = []
        while len(rawlist) >= 3:
            if rawlist[1] == 0x2d:  # rawlist[0] '-' rawlist[2]
                a, b = rawlist[0], rawlist[2]
                if a > b:
                    # TODO: complain
                    a, b = b, a  # swap
                usedlist.extend(range(a, b + 1))
                rawlist = rawlist[3:]
            else:
                a = rawlist.pop(0)
                usedlist.append(a)

        # At this point, there are at most two items left.
        usedlist.extend(rawlist)

        # Is the list unequal to the set: complain.
        usedset = set(usedlist)
        if len(usedset) != len(usedlist):
            # print('TODO/FIXME/COMPLAIN:', usedset, usedlist)
            pass

        # Numeric value.
        usedlist = list(usedset)
        usedlist.sort()
        num = len(usedlist) * 0x100 | usedlist[0]

        # Check if usedlist has gaps, if it has, then return a tuple
        # instead of a single value.
        if len(usedlist) > 1 and (
                usedlist[-1] - usedlist[0] != len(usedlist) - 1):
            return (num, bytes(usedlist))

        return (num,)

    @property
    def is_canonical(self):
        if not hasattr(self, '_is_canonical'):
            # Strip dashes from the raw version.
            raw_no_dash = ''.join(self.RE_NO_DASH.findall(self.raw))
            # Compare the dashless version with the canonical one.
            self._is_canonical = (raw_no_dash == self.canonical_pattern)
        return self._is_canonical

    @property
    def canonical_pattern(self):
        if not hasattr(self, '_canonical_pattern'):
            if self.values[0] == self.NOT_A_PATTERN:
                # Hrm.. latin1 here.. not so nice.
                ret = bytes((i[0] & 0xff)
                            for i in self.values[1:]).decode('latin1')
            # Don't trust IS_A_PATTERN; it may not be a pattern after
            # all. (TODO: decide whether we can simply drop the pattern,
            # since it does affect sort order :-/ )
            elif (all(i[0] < 0x200 for i in self.values[1:-1]) and
                    self.values[-1] == (0x30000,)):
                ret = bytes((i[0] & 0xff)
                            for i in self.values[1:-1]).decode('latin1')
            else:
                # Okay, so we have a pattern with pattern stuff. Fix it
                # up.
                assert self.values[0] == self.IS_A_PATTERN
                ret = ('_' + self._canonical_pattern_range(
                    self.values[1:-1]).decode('latin1'))
            self._canonical_pattern = ret
        return self._canonical_pattern

    def _canonical_pattern_range(self, values):
        common = {
            0xa30: 0x58,    # 'X'
            0x931: 0x5a,    # 'Z'
            0x832: 0x4e,    # 'N'
            0x18000: 0x2e,  # '.'
            0x28000: 0x21,  # '!'
        }
        special = set([0x21, 0x2d, 0x2e,  # '!', '-', '.'
                       0x4e, 0x58, 0x5a, 0x6e, 0x78, 0x7a])  # NnXxZz

        ret = []
        for value in values:
            if len(value) == 1:
                num = value[0]
                if num < 0x200:
                    # Single character.
                    start = num & 0xff
                    if start in special:  # must escape these in range
                        ret.extend([0x5b, start, 0x5d])
                    else:
                        ret.append(num & 0xff)
                else:
                    # Easy range.
                    try:
                        ret.append(common[num])
                    except KeyError:
                        length = (num & 0xff00) >> 8
                        start = num & 0xff
                        if length == 2:
                            ret.extend([0x5b, start, start + 1, 0x5d])
                        elif length == 3:
                            ret.extend([0x5b, start, start + 1, start + 2,
                                        0x5d])
                        else:
                            ret.extend([0x5b, start, 0x2d,
                                        start + length - 1, 0x5d])
            else:
                # Complex range.
                self._canonical_pattern_add_range(ret, list(value[1]))

        return bytes(ret)

    def _canonical_pattern_add_range(self, ret, binstr):
        assert binstr
        ret.append(0x5b)
        dash = (0x2d in binstr)
        if dash:
            binstr.remove(0x2d)
        if (0x5d in binstr):
            binstr.remove(0x5d)
            ret.append(0x5d)
        ranges = self._canonical_pattern_get_ranges(binstr)
        # Reorder ranges: first numeric and alpha, and then the rest.
        # (Does this cause trouble if the dash is in a range?)
        ranges.sort(key=self._canonical_pattern_range_sort)
        for range_ in ranges:
            if len(range_) == 1:
                ret.append(range_[0])
            else:
                assert len(range_) == 2
                length = range_[1] - range_[0] + 1
                if length <= 3:
                    ret.extend([range_[0] + i for i in range(length)])
                else:
                    ret.extend([range_[0], 0x2d, range_[1]])
        if dash:
            ret.append(0x2d)
        ret.append(0x5d)

    def _canonical_pattern_get_ranges(self, binstr):
        ranges = []
        while binstr:
            maxi = len(binstr) - 1
            i = 1
            while maxi >= 2 and i <= maxi:
                if (binstr[i] - binstr[0] != i):
                    break
                i += 1
            i -= 1
            if 2 <= i <= maxi and (binstr[i] - binstr[0] == i):
                ranges.append((binstr[0], binstr[i]))
                binstr[0:i + 1] = []  # inline remove
            else:
                ranges.append((binstr.pop(0),))
        return ranges

    @staticmethod
    def _canonical_pattern_range_sort(range_):
        # Fix so 0-9A-Za-z comes first.
        if (0x30 <= range_[0] <= 0x39 or
                0x41 <= range_[0] <= 0x5a or
                0x61 <= range_[0] <= 0x7a):
            return range_[0]
        return range_[0] + 0x100

    def matches_same(self, other):
        """
        Same as regular equal, but this time, ignore whether this is
        a pattern or not.
        """
        if other is None:
            return False
        if self.values[0] == other.values[0]:  # both pattern or both not
            return self.values == other.values
        if self.values[0] == self.IS_A_PATTERN:
            return self.values[1:-1] == other.values[1:]    # self is pattern
        return self.values[1:] == other.values[1:-1]        # other is pattern

    def __hash__(self):
        return hash(self.values)

    def __eq__(self, other):
        if other is None:
            return False
        return self.values == other.values

    def __lt__(self, other):
        if other is None:
            return False
        return self.values < other.values

    def __repr__(self):
        return '<Pattern({})>'.format(self.raw or '')

    def __str__(self):
        return self.raw



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import AppBase, IfStyleApp

# In the following commit -- between Asterisk 1.4 and Asterisk 1.6.x --
# not only did the default App argument delimiter change from pipe (|)
# to comma (,), but also the ExecIf change to GotoIf-style syntax, using
# a question mark (?).
#
# > commit 55b1ee298e926c594f45456e5afbc25c79e3889b
# > Author: Tilghman Lesher <tilghman@CENSORED>
# > Date:   Mon Jul 23 19:51:41 2007 +0000
#
# > Merge the dialplan_aesthetics branch.  Most of this patch simply
# > converts applications using old methods of parsing arguments to
# > using the standard macros.  However, the big change is that the
# > really old way of specifying application and arguments separated by
# > a comma will no longer work (e.g. NoOp,foo|bar).  Instead, the way
# > that has been recommended since long before 1.0 will become the only
# > method available (e.g. NoOp(foo,bar).
# >
# > git-svn-id: https://origsvn.digium.com/svn/asterisk/trunk@76703
# >   65c4cc65-6c06-0410-ace0-fbb531ad65f3


class Exec(AppBase):
    pass


class ExecIf(IfStyleApp):
    pass


class TryExec(AppBase):
    pass


def register(app_loader):
    for app in (
            Exec, ExecIf, TryExec):
        app_loader.register(app())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Show dialplan like Asterisk does with CLI command "dialplan show". Takes
'extensions.conf' as argument.
"""
import argparse

from asterisklint import FileDialplanParser
from asterisklint.defines import MessageDefManager


def main(args, envs):
    parser = argparse.ArgumentParser(
        description=(
            'Shows the dialplan like Asterisk does with the CLI command '
            '"dialplan show". Useful for testing whether asterisklint '
            'parser the input properly.'))
    parser.add_argument(
        'dialplan', metavar='EXTENSIONS_CONF',
        help="path to extensions.conf")
    parser.add_argument(
        '--reverse', action='store_true',
        help="some versions of Asterisk output the dialplan file in reverse")
    args = parser.parse_args(args)

    MessageDefManager.muted = True  # no messages to stderr
    parser = FileDialplanParser()
    parser.include(args.dialplan)
    dialplan = next(iter(parser))
    print(dialplan.format_as_dialplan_show(
        reverse=args.reverse))



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .defines import ErrorDef, WarningDef
from .where import Where


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_FILE_UTF8_BAD(ErrorDef):
        message = 'expected UTF-8 encoding, got something else'

    class W_FILE_CTRL_CHAR(WarningDef):
        message = 'unexpected control character found'

    class W_FILE_DOS_EOFCRLF(WarningDef):
        message = 'unexpected trailing CRLF in DOS file format'

    class W_FILE_DOS_BARELF(WarningDef):
        message = 'unexpected bare LF in DOS file format'

    class W_FILE_UNIX_CRLF(WarningDef):
        message = 'unexpected CRLF in UNIX file format'

    class W_FILE_UNIX_NOLF(WarningDef):
        message = 'unexpected line without LF in UNIX file format'

    class W_WSH_BOL(WarningDef):
        message = 'unexpected leading whitespace'

    class W_WSH_EOL(WarningDef):
        message = 'unexpected trailing whitespace'

    class W_WSH_COMMENT(WarningDef):
        message = 'missing whitespace around comment semicolon'

    class W_WSV_BOF(WarningDef):
        message = 'unexpected vertical space at beginning of file'

    class W_WSV_EOF(WarningDef):
        message = 'unexpected vertical space at end of file'


class BinFileReader(object):
    """
    Reads the binary opened file fp, but may open more files with the
    opener if an #include directive is encountered.
    """
    def __init__(self, opener=(lambda filename: open(filename, 'rb'))):
        self._files = []
        self._generators = []
        self._filenames = []

        self._opener = opener

    def include(self, filename):
        fp = self._opener(filename)

        if hasattr(fp, 'mode'):
            assert 'b' in fp.mode, 'expected binary opened file'

        self._files.append(fp)
        self._generators.append(enumerate(fp))
        self._filenames.append(fp.name)

    def __iter__(self):
        while self._generators:
            try:
                i, line = next(self._generators[-1])
            except StopIteration:
                if hasattr(self._files[-1], 'close'):
                    self._files[-1].close()
                self._files.pop()
                self._generators.pop()
                self._filenames.pop()
            else:
                yield Where(self._filenames[-1], i + 1, line), line


class EncodingReader(object):
    """
    Decodes lines from UTF-8, the one true encoding.
    """
    def __iter__(self):
        for where, data in super(EncodingReader, self).__iter__():
            try:
                data = data.decode('utf-8')
            except UnicodeDecodeError:
                E_FILE_UTF8_BAD(where)
                data = data.decode('cp1252')  # or latin1? or 9?
            yield where, data


class NoCtrlReader(object):
    """
    Checks for unusual control characters.
    """
    # all, except \r (0d), \n (0a), \t (09)
    illegal = set('\x00\x01\x02\x03\x04\x05\x06\x07'
                  '\x08'      '\x0b\x0c'  '\x0e\x0f'
                  '\x10\x11\x12\x13\x14\x15\x16\x17'
                  '\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f')

    def __iter__(self):
        for where, data in super(NoCtrlReader, self).__iter__():
            charset = set(data)
            union = charset & self.illegal
            if union:
                W_FILE_CTRL_CHAR(where)
            yield where, data


class FileformatReader(object):
    """
    TODO: allow one to specify legal line endings (unix vs dos)
    """
    def _pop_fileinfo(self, fileinfo, stop_at):
        while fileinfo and fileinfo[-1][0] != stop_at:
            fn, is_dos, latest_has_lf, last_where = fileinfo.pop()
            assert None not in (is_dos, latest_has_lf)
            if is_dos and latest_has_lf:
                W_FILE_DOS_EOFCRLF(last_where)
            elif not is_dos and not latest_has_lf:
                W_FILE_UNIX_NOLF(last_where)

    def __iter__(self):
        # We store fileinfo in a stack so we can look backwards from
        # includes, so check file endings.
        fileinfo = []  # [[filename, is_dos], ...]

        for where, data in super(FileformatReader, self).__iter__():
            # Keep track of includes.
            if not fileinfo or fileinfo[-1][0] != where.filename:
                # Where we in here already? Then we're backing out of an
                # include.
                if where.filename in [i[0] for i in fileinfo]:
                    self._pop_fileinfo(fileinfo, where.filename)
                # We weren't? Then we're moving into an include.
                else:
                    fileinfo.append([where.filename, None, None, where])
                # Re-set values.
                filename, is_dos, latest_has_lf, last_where = fileinfo[-1]

            has_crlf = data.endswith('\r\n')
            has_lf = data.endswith('\n')
            fileinfo[-1][2] = has_lf
            fileinfo[-1][3] = where

            if is_dos is None:
                is_dos = (has_crlf or not has_lf)
                fileinfo[-1][1] = is_dos

            if is_dos:
                if has_lf and not has_crlf:
                    W_FILE_DOS_BARELF(where)
            else:
                if has_crlf:
                    W_FILE_UNIX_CRLF(where)

            if has_crlf:
                data = data[0:-2]
            elif has_lf:
                data = data[0:-1]

            yield where, data

        self._pop_fileinfo(fileinfo, None)


class AsteriskCommentReader(object):
    """
    Unescapes backslash escapes and splits the data from the comments.

    TODO: look at: main/config.c: process_text_line()
    TODO: also parse multiline asterisk comments
    """
    @staticmethod
    def simple_comment_split(data, where):
        try:
            i = data.index(';')
        except ValueError:
            return data, ''
        if i > 0 and data[i - 1] not in ' \t':
            W_WSH_COMMENT(where)
        while i > 0 and data[i - 1] in ' \t':
            i -= 1
        return data[0:i], data[i:]

    def __iter__(self):
        for where, data in super(AsteriskCommentReader, self).__iter__():
            # We cannot escape whitespace, so no need to keep this
            # around.
            if data.endswith(tuple(' \t')):
                W_WSH_EOL(where)
                data = data.rstrip(' \t')

            # Shortcut if we don't do any escaping.
            if '\\' not in data:
                data, comment = self.simple_comment_split(data, where)
                yield where, data, comment
                continue

            # Asterisk does really poor backslash escaping in the config
            # decoder. What it does amounts to a lookback only:
            # - is it a semi? check previous char for backslash
            # - if no backslash, break here
            # - if backslash, replace both with single semi and continue
            i = 0
            parts = []
            while True:
                try:
                    i = data.index(';')
                except ValueError:
                    parts.append(data)
                    comment = ''
                    break
                else:
                    if i and data[i - 1] == '\\':
                        parts.append(data[0:i - 1] + ';')
                        data = data[i + 1:]
                    else:
                        parts.append(data[0:i])
                        comment = data[i:]
                        break
            data = ''.join(parts)

            # Move all the whitespace at the end of data to comment.
            i = len(data)
            if comment and i > 0 and data[i - 1] not in ' \t':
                W_WSH_COMMENT(where)
            while i > 0 and data[i - 1] in ' \t':
                i -= 1
            comment = data[i:] + comment
            data = data[0:i]

            yield where, data, comment


class FileReader(AsteriskCommentReader, FileformatReader, NoCtrlReader,
                 EncodingReader, BinFileReader):
    pass



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .application import App
from .config import ConfigAggregator, Context, Varset
from .config import E_CONF_CTX_MISSING, E_CONF_KEY_INVALID, W_CONF_CTX_DUPE
from .defines import ErrorDef, WarningDef, HintDef, DupeDefMixin
from .pattern import H_PAT_NON_CANONICAL, Pattern
from .where import Where


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class I_NOTIMPL_HINT(ErrorDef):
        message = 'the dialplan hint directive has not been implemented yet'

    class I_NOTIMPL_IGNOREPAT(ErrorDef):
        message = ('the dialplan ignorepat directive has not been implemented '
                   'yet')

    class I_NOTIMPL_SWITCH(ErrorDef):
        message = 'the dialplan switch directive has not been implemented yet'

    class E_DP_VAR_INVALID(ErrorDef):
        message = ('unexpected variable name {variable!r} '
                   '(not exten/same/include)')

    class E_DP_INCLUDE_POSITION(WarningDef):
        message = 'an include should always be at the tail of the context'

    class E_DP_PAT_INVALID(ErrorDef):
        message = 'badly formatted pattern'

    class E_DP_PAT_MISSING(ErrorDef):
        message = 'missing pattern'

    class E_DP_PRIO_INVALID(ErrorDef):
        message = 'invalid priority {prio!r}'

    class E_DP_PRIO_MISSING(ErrorDef):
        message = 'missing priority'

    class E_DP_PRIO_DUPE(DupeDefMixin, ErrorDef):
        message = 'duplicate priority {prio!r}'

    class E_DP_LABEL_INVALID(ErrorDef):
        message = 'unexpected label value {label!r}'

    class E_DP_GLOBALS_DUPE(DupeDefMixin, ErrorDef):
        message = 'second globals context, will not be used'

    class E_DP_LABEL_DUPE(WarningDef):
        message = 'duplicate label {label!r}, one of them will not be used'

    class E_DP_PRIO_BAD(ErrorDef):
        message = 'cannot match priority for pattern {pat!r}'

    class W_DP_PRIO_BADORDER(WarningDef):
        message = 'bad priority order for pattern {pat!r} and prio {prio}'

    class H_DP_GENERAL_MISPLACED(HintDef):
        message = '[general] context not found or not at top of file'

    class H_DP_GLOBALS_MISPLACED(HintDef):
        message = '[globals] context not found or not at position two'


class Dialplan(object):
    def __init__(self):
        self._general = None
        self._globals = None
        self.contexts = []
        self.last_prio = None  # used by the DialplanContext

    @property
    def general(self):
        """If it doesn't exist, we return an empty object, so it behaves
        normally."""
        return self._general or DialplanContext('general', dialplan=self)

    @property
    def globals(self):
        """If it doesn't exist, we return an empty object, so it behaves
        normally."""
        return self._globals or DialplanContext('globals', dialplan=self)

    def add_general(self, general):
        if self._general:
            W_CONF_CTX_DUPE(general.where, previous=self._general.where)
            # BEWARE: make sure we use a ref to the NEW context, because
            # we may be appending to that object...!!
            self._general.extend(general)
        else:
            if self._globals or self.contexts:
                H_DP_GENERAL_MISPLACED(general.where)
            self._general = general

    def add_globals(self, globals_):
        if self._globals:
            E_DP_GLOBALS_DUPE(globals_.where, previous=self._globals.where)
            # don't save
        else:
            if not self._general or self.contexts:
                H_DP_GLOBALS_MISPLACED(globals_.where)
            self._globals = globals_

    def get_where(self):
        where = None
        if self.contexts:
            where = Where(
                filename=self.contexts[0].where.filename,
                lineno=0, line='')
        return where

    def on_complete(self):
        if not self._general:
            H_DP_GENERAL_MISPLACED(self.get_where())
        if not self._globals:
            H_DP_GLOBALS_MISPLACED(self.get_where())

    def format_as_dialplan_show(self, reverse=True):
        # Okay; please explain the reverse madness to me? I've seen
        # output from one source in proper order, but usually (Asterisk
        # 13, Asterisk 1.4) it comes in reversed order...
        #
        # Use the `dialplan show` to compare with asterisk output.
        contexts = self.contexts
        if reverse:
            contexts = reversed(contexts)

        ret = []
        for context in contexts:
            ret.append("[ Context {!r} created by 'pbx_config' ]".format(
                context.name))
            last_pattern = None
            for extension in context.by_pattern():
                # NOTE: Asterisk 1.4 and LOW_MEMORY limits show_dialplan
                # output of extension.prio+app_with_parents to 255
                # chars. Asterisk 11 limits it to 1023.
                if extension.pattern == last_pattern:
                    ret.append('     %-14s %-45s [pbx_config]' % (
                        (extension.label and '[{}]'.format(extension.label) or
                            ''),
                        ('%d. %s' % (extension.prio,
                                     extension.app_with_parens))))
                else:
                    last_pattern = extension.pattern
                    ret.append('  %-17s %-45s [pbx_config]' % (
                        "'%s' =>" % (extension.pattern,),
                        ('%d. %s' % (extension.prio,
                                     extension.app_with_parens))))
            for include in context.includes:
                ret.append('  %-17s %-45s [pbx_config]' % (
                    'Include =>', "'%s'" % (include.context_name,)))
            ret.append('')
        return '\n'.join(ret)


class DialplanContext(Context):
    def __init__(self, *args, **kwargs):
        self.dialplan = kwargs.pop('dialplan')
        super().__init__(*args, **kwargs)
        self.includes = []
        self.context_last_prio = None

    def update(self, othercontext):
        assert not othercontext.includes
        assert othercontext.context_last_prio is None
        self.context_last_prio = None
        super().update(othercontext)

    def by_pattern(self):
        """
        Used by format_as_dialplan_show().
        """
        patterns = sorted(list(set(i.pattern for i in self)))
        ret = []
        for pattern in patterns:
            ret.extend(i for i in self if i.pattern == pattern)
        return ret

    def add(self, extension):
        # - If extension.pattern is None ("same") then take previous.
        #   Error if there is none.
        # - If prio is N then assert there is a previous prio with same
        #   pattern.
        if extension.pattern is None:
            if not len(self):
                E_DP_PAT_MISSING(extension.where)
                return None
            extension.pattern = self[-1].pattern

        if extension.prio is None:
            if not self.dialplan.last_prio:
                # Can't use 'next' priority on the first entry at line
                # ..  of extensions.conf!
                E_DP_PRIO_MISSING(
                    extension.where, pat=extension.pattern.raw)
                return
            elif (not self.context_last_prio or
                    self[-1].pattern != extension.pattern):
                W_DP_PRIO_BADORDER(
                    extension.where, pat=extension.pattern.raw,
                    prio='<unset>')
            extension.prio = self.dialplan.last_prio + 1
        elif extension.prio < 1:
            E_DP_PRIO_INVALID(extension.where, prio=extension.prio)
            return
        elif extension.prio != 1:
            # Check that there is a prio with N-1.
            try:
                prev = [i for i in self if i.pattern == extension.pattern][-1]
            except IndexError:
                # TODO: here we have to be careful with pattern
                # matching, we should check whether a pattern exists
                # with a greater scope than our pattern.
                # The odd pattern order can be succesfully employed
                # in temporary branching situations:
                #   _X!,1,both
                #   _1!,2,specific-1-stuff
                #   _[02-9]!,2,specific-other-stuff
                #   _X!,3,both
                W_DP_PRIO_BADORDER(
                    extension.where, pat=extension.pattern.raw,
                    prio=extension.prio)
            else:
                if prev.prio != extension.prio - 1:
                    # Don't warn if same prio, we do the dupe check
                    # later on.
                    if prev.prio != extension.prio:
                        W_DP_PRIO_BADORDER(
                            extension.where, pat=extension.pattern.raw,
                            prio=extension.prio)

        # Okay, prio is valid here. Store it so we can use the next 'n'
        # as last_prio + 1.
        self.dialplan.last_prio = extension.prio
        self.context_last_prio = extension.prio

        # Check duplicate priorities.
        extensions = [i for i in self if i.pattern == extension.pattern]
        if any(i.prio == extension.prio for i in extensions):
            previous = [i.where for i in extensions
                        if i.prio == extension.prio][0]
            E_DP_PRIO_DUPE(
                extension.where, previous=previous, prio=extension.prio)
            return

        # Check duplicate labels.
        if (extension.label and
                any(True for i in extensions if i.label == extension.label)):
            previous = [i.where for i in extensions
                        if i.label == extension.label][0]
            E_DP_LABEL_DUPE(
                extension.where, previous=previous, label=extension.label)
            extension.label = ''  # wipe it

        super(DialplanContext, self).add(extension)

    def add_include(self, include):
        assert isinstance(include, Include)
        # FIXME: check for dupes and other stupidity (circular includes?)
        # FIXME: complain when the include is not at the tail of the context
        self.includes.append(include)


class DialplanVarset(object):
    @classmethod
    def from_varset(cls, varset):
        """
        Use this on subclasses of Varset.
        """
        if varset.variable in ('exten', 'same'):
            assert varset.arrow  # W_ARROW
            if varset.variable == 'exten':
                pattern, rest = varset.value.split(',', 1)
            else:
                pattern, rest = None, varset.value

            try:
                prio, app = rest.split(',', 1)
            except ValueError:
                prio, app = rest, ''
            try:
                prio, label = prio.split('(', 1)
            except ValueError:
                prio, label = prio, None
            if prio == 'n':
                prio = None
            elif prio.isdigit():
                prio = int(prio)
            elif prio == 'hint':
                I_NOTIMPL_HINT(varset.where)
                return None
            else:
                E_DP_PRIO_INVALID(varset.where, prio=prio)
                return None

            if label is not None:
                if not label.endswith(')'):
                    E_DP_LABEL_INVALID(varset.where)
                    return None
                label = label[0:-1]
                # TODO: check label validity
            else:
                label = None

            pattern = pattern and Pattern(pattern, varset.where) or None
            app = App(app, varset.where)
            return Extension(pattern, prio, label, app, varset.comment,
                             varset.where)

        elif varset.variable == 'include':
            assert varset.arrow  # W_ARROW
            return Include(varset.value, varset.where)

        elif varset.variable == 'switch':
            assert varset.arrow  # W_ARROW
            I_NOTIMPL_SWITCH(varset.where)

        elif varset.variable == 'ignorepat':
            assert varset.arrow  # W_ARROW
            I_NOTIMPL_IGNOREPAT(varset.where)

        else:
            E_CONF_KEY_INVALID(varset.where, key=varset.variable)
            return None


class Include(object):
    def __init__(self, value, where):
        # FIXME: check for value name? contexts existence?
        self.context_name = value
        self.where = where


class Extension(Varset):
    def __init__(self, pattern, prio, label, app, comment, where):
        # Check pattern voor mixen van letters en pattern-letters:
        # - "s-zap" == "s-[1-9]ap" en zou als "s-[z]ap" geschreven moeten
        # - "s-abc" == "sabc"
        # Als pattern is None, dan pakt Context gewoon de vorige, da's
        # prima.
        self.pattern = pattern
        self.prio = prio
        self.label = label
        self.app = app
        self.comment = comment
        self.where = where

    @property
    def app_with_parens(self):
        """
        Used by format_as_dialplan_show().
        """
        ret = [self.app.raw]
        if '(' not in self.app.raw:
            ret.append('(')
        if ')' not in self.app.raw:
            ret.append(')')
        return ''.join(ret)

    @property
    def prio_with_label(self):
        if self.label:
            return '{}({})'.format(self.prio, self.label)
        return str(self.prio)

    def __repr__(self):
        return '<Extension({},{})>'.format(
            self.pattern.raw, self.prio_with_label)


class DialplanAggregator(ConfigAggregator):
    def on_begin(self):
        self._dialplan = Dialplan()
        self._prevcontexts = {}
        self._curcontext = None

    def on_yield(self):
        self._dialplan.on_complete()
        yield self._dialplan

    def on_context(self, context):
        if context.name == 'general':
            self._dialplan.add_general(context)
            self._curcontext = context
        elif context.name == 'globals':
            self._dialplan.add_globals(context)
            self._curcontext = context
        else:
            dialplan_context = DialplanContext.from_context(
                context, dialplan=self._dialplan)
            if dialplan_context:
                self.on_dialplancontext(dialplan_context)

    def on_varset(self, varset):
        if not self._curcontext:
            E_CONF_CTX_MISSING(varset.where)
        elif isinstance(self._curcontext, DialplanContext):
            dialplanvarset = DialplanVarset.from_varset(varset)
            if dialplanvarset:
                self.on_dialplanvarset(dialplanvarset)
        else:
            self._curcontext.add(varset)

    def on_dialplancontext(self, dialplancontext):
        oldcontext = self._prevcontexts.get(dialplancontext.name)
        if oldcontext:
            oldcontext.update(dialplancontext)
            self._curcontext = oldcontext
        else:
            self._prevcontexts[dialplancontext.name] = dialplancontext
            self._dialplan.contexts.append(dialplancontext)
            self._curcontext = dialplancontext

    def on_dialplanvarset(self, dialplanvarset):
        assert isinstance(self._curcontext, DialplanContext)
        if isinstance(dialplanvarset, Extension):
            pattern = dialplanvarset.pattern
            if pattern and not pattern.is_canonical:
                H_PAT_NON_CANONICAL(
                    dialplanvarset.where, pat=pattern.raw,
                    expected=pattern.canonical_pattern)
            self._curcontext.add(dialplanvarset)
        elif isinstance(dialplanvarset, Include):
            self._curcontext.add_include(dialplanvarset)
        else:
            raise NotImplementedError()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .variable import Var


class Expr(Var):
    """
    A special case of Var where an expression is evaluated.

    TODO: complain about (expr);
    - excess whitespace
    - no agreement on quotes on either side of expression
    - bad/unknown operators
    - joined expressions (expr in expr, which is not needed, because you
      can use parens)
    """
    def __init__(self, expression=None):  # drop start and length args
        super().__init__(name=expression)

    def __str__(self):
        if self.name:
            return '$[{}]'.format(self.name)
        return super().__str__()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..base import AppBase


class ConfBridge(AppBase):
    pass


def register(app_loader):
    app_loader.register(ConfBridge())



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Var(object):
    """
    Var is a nested structure of variables.

    For example::

        "${abc}"    => Var('abc')
        "abc${def}" => Var.join(['abc', Var('abc')])
        "${${abc}}" => Var(Var('abc'))
        "${a${bc}}" => Var(Var.join(['a', Var('bc')]))

    You can call stringify the Var back to their original values::

        str(Var('abc')) => "${abc}"
        ...

    You can also format the Var with variable replacements:

        Var(Var.join(['a', Var('bc')])).format(bc='BC', aBC='foobar') =>
          "foobar"
    """
    @classmethod
    def join(self, variables):
        copy = []
        for var in variables:
            if not isinstance(var, (str, Var)):
                raise TypeError(
                    'Variables can only consist of strings and other '
                    'variables, got {!r}'.format(var))
            if var:
                copy.append(var)

        if len(copy) == 0:
            return ''
        if len(copy) == 1:
            return copy[0]

        # Use private setter instead of constructor.
        ret = Var()
        ret._list = copy
        return ret

    def __init__(self, name=None, start=None, length=None):
        self.name = name

    def format(self, **kwargs):
        if self.name is None:
            ret = []
            for var in self._list:
                if isinstance(var, str):
                    ret.append(var)
                else:
                    ret.append(var.format(**kwargs))
            return ''.join(ret)

        elif isinstance(self.name, Var):
            ret = self.name.format(**kwargs)
            return kwargs[ret]

        else:
            return kwargs[self.name]

    def __iter__(self):
        """
        You may iterate over this thing, as if it where a string. In
        which case you get the literal letters, and variables in
        between.
        """
        return iter(self._get_cached_iter())

    def __getitem__(self, *args, **kwargs):
        """
        Get an item or list of items.

        If a slice is requested (and a list is returned), we attempt to
        join the list using strjoin. Examples::

            ['A', 'B', 'C', <Var>, 'D'] ==> ['ABC', <Var>, 'D']
            ['A', 'B', 'C'] ==> 'ABC'
            ['A'] ==> 'A'
            [<Var>] ==> <Var>
            [] ==> []

        This should never be a problem because you'll be iterating over
        the results anyway.
        """
        item = self._get_cached_iter().__getitem__(*args, **kwargs)
        if isinstance(item, list):
            ret = list(strjoin(item))  # joins consecutive strings together
            if len(ret) == 1:
                return ret[0]
            return ret
        return item

    def __len__(self):
        return len(self._get_cached_iter())

    def _get_cached_iter(self):
        if not hasattr(self, '_cached_iter'):
            ret = []
            if self.name is None:
                for var in self._list:
                    for inner in var:
                        ret.append(inner)
            else:
                ret.append(self)
            self._cached_iter = ret
        return self._cached_iter

    def __eq__(self, other):
        if not isinstance(other, Var):
            return False

        if (self.name, other.name) == (None, None):
            if len(self._list) != len(other._list):
                return False
            return all(self._list[i] == other._list[i]
                       for i in range(len(self._list)))

        if None in (self.name, other.name):
            return False

        return self.name == other.name

    def __str__(self):
        if self.name:
            return '${{{}}}'.format(self.name)
        return ''.join(str(i) for i in self._list)

    def split(self, token):
        if token != ':':
            raise TypeError('expected split on colon, got {!r}'.format(token))

        assert not self.name
        assert len(self._list) >= 2
        ret = self._list[0].split(token)
        assert len(ret) > 1, ret

        # Split it up by token, so we group the values into subgroups.
        split_up = [[]]
        for item in self._list:
            if isinstance(item, str):
                subitems = item.split(token)
                while len(subitems) > 1:
                    first = subitems.pop(0)
                    if first:
                        split_up[-1].append(first)
                    split_up.append([])
                if subitems[0]:
                    split_up[-1].append(subitems[0])
            else:
                split_up[-1].append(item)

        # Join the subgroups back into variables.
        for i, list_ in enumerate(split_up):
            assert len(list_)
            if len(list_) > 1:
                split_up[i] = Var.join(split_up[i])
            else:
                split_up[i] = split_up[i][0]

        return split_up


class SliceMixin(object):
    def __init__(self, *args, start=None, length=None, **kwargs):
        assert start is not None
        super().__init__(*args, **kwargs)
        self.start = start
        self.length = length
        if length is not None:
            assert length != 0
            if length < 0:
                self._endpos = length
            elif self.start < 0:
                self._endpos = self.start + length
                assert self._endpos < 0
            else:
                self._endpos = self.start + length

    def format(self, **kwargs):
        value = super().format(**kwargs)
        if self.length:
            return value[self.start:self._endpos]
        return value[self.start:]

    def __str__(self):
        if self.length:
            return '${{{}:{}:{}}}'.format(
                self.name, self.start, self.length)
        return '${{{}:{}}}'.format(
            self.name, self.start)


class VarSlice(SliceMixin, Var):
    def __init__(self, name=None, start=None, length=None):
        assert name is not None
        super().__init__(name=name, start=start, length=length)


class VarDynSlice(Var):
    def __init__(self, name=None, start=None, length=None):
        assert name is not None
        super().__init__(name=name)
        self.start = start
        self.length = length

    def __str__(self):
        if self.length:
            return '${{{}:{}:{}}}'.format(
                self.name, self.start, self.length)
        return '${{{}:{}}}'.format(
            self.name, self.start)


def strjoin(list_of_items_and_strings):
    """
    Joins all consecutive items that are strings together.

    E.g.: [1, 2, 'a', 'b', 'c', 3, 'd', 'e'] ==> [1, 2, 'abc', 3, 'de']
    """
    cache = []
    for ch in list_of_items_and_strings:
        if isinstance(ch, str):
            cache.append(ch)
        else:
            if cache:
                yield ''.join(cache)
                cache = []
            yield ch
    if cache:
        yield ''.join(cache)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys
from collections import defaultdict


class MessageDefException(Exception):
    pass


class DuplicateMessageDef(MessageDefException):
    pass


class MessageDefManager(type):
    types = set()
    muted = False
    mute_by_message = set()

    show_line = False
    show_previous = False

    raised = defaultdict(list)

    # The metaclass invocation. (No need for __prepare__ at this
    # point.)
    def __new__(cls, name, bases, classdict):
        class_mute = False

        if name in cls.types:
            raise DuplicateMessageDef(name)
        if not name.endswith('Def'):  # MessageDef, ErrorDef, WarningDef, ...
            cls.types.add(name)       # W_FOO, E_BAR

            # We want the user to specify items to ignore somehow. For
            # now we'll use the environment variable ALINT_IGNORE with a
            # comma delimited list of messages to silence.
            # Example: export ALINT_IGNORE=E_APP_ARG_IFSTYLE,E_FUNC_MISSING,W_
            # (Will silence *all* warnings, through the "W_".)
            alint_ignore = tuple(
                i for i in set(os.environ.get('ALINT_IGNORE', '').split(','))
                if i)
            if name.startswith(alint_ignore):
                for ignore in alint_ignore:
                    if (name == ignore or
                            (ignore[-1] == '_' and
                             name.startswith(ignore))):
                        class_mute = True
                        break

        class_ = type.__new__(cls, name, bases, classdict)

        # Add list of callbacks, specific per class.
        class_._callbacks = []

        # Add class specific mute setting.
        class_.muted = class_mute

        return class_

    @classmethod
    def reset(cls):
        cls.raised = defaultdict(list)

    @classmethod
    def on_message(cls, msg):
        cls.raised[msg.__class__.__name__].append(msg)
        formatted = msg.message.format(**msg.fmtkwargs)

        if cls.muted or msg.muted:
            return

        print('{} {}: {}'.format(
            msg.where, msg.__class__.__name__, formatted),
            file=sys.stderr)

        if cls.show_line:
            line_repr = repr(msg.where.line)
            if len(line_repr) > 64:
                line_repr = line_repr[1:64] + '...'  # drop "b", trim to 63
            else:
                line_repr = line_repr[1:]  # drop "b"
            print('  => {}'.format(line_repr), file=sys.stderr)

        if cls.show_previous and msg.previous:
            print(
                '    previous occurrence at {}'.format(msg.previous),
                file=sys.stderr)


class MessageDef(object, metaclass=MessageDefManager):
    @classmethod
    def add_callback(cls, callback):
        # self._callbacks is subclass-specific, initialized through our
        # metaclass.
        cls._callbacks.append(callback)

    @classmethod
    def call_callbacks(cls, instance):
        for callback in cls._callbacks:
            callback(instance)

    def __init__(self, where, previous=None, **fmtkwargs):
        self.where = where
        self.previous = previous
        self.fmtkwargs = fmtkwargs
        # Notify optional callbacks.
        self.call_callbacks(self)
        # Notify our general manager.
        MessageDefManager.on_message(self)


class ErrorDef(MessageDef):
    pass


class WarningDef(MessageDef):
    pass


class HintDef(MessageDef):
    pass


class DupeDefMixin(object):
    def __init__(self, where, **kwargs):
        if not kwargs.get('previous'):
            raise TypeError("{} requires a ``previous'' argument".format(
                self.__class__.__name__))
        super(DupeDefMixin, self).__init__(where, **kwargs)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from distutils.core import setup
from os import walk
from os.path import abspath, dirname, join


def get_packages():
    here = dirname(abspath(__file__))
    superdir = join(here, 'asterisklint')
    packages = []
    for root, dirs, files in walk(superdir):
        packages.append(root[len(here) + 1:].replace('/', '.'))
        if '__pycache__' in dirs:
            dirs.remove('__pycache__')
    packages.sort()
    return packages


if __name__ == '__main__':
    if sys.version_info < (3,):
        raise NotImplementedError(
            "I'm sorry, but asterisklint works with Python3+ only")

    long_descriptions = []
    with open(join(dirname(__file__), 'README.rst')) as file:
        long_descriptions.append(file.read())
    with open(join(dirname(__file__), 'CHANGES.rst')) as file:
        long_descriptions.append(file.read())
    version = long_descriptions[-1].strip().split(' ', 1)[0]

    setup(
        name='asterisklint',
        version=version,
        scripts=['scripts/asterisklint'],
        packages=get_packages(),
        data_files=[('', ['CHANGES.rst', 'LICENSE', 'README.rst'])],
        description='Asterisk PBX configuration syntax checker',
        long_description=('\n\n\n'.join(long_descriptions)),
        author='Walter Doekes, OSSO B.V.',
        author_email='wjdoekes+asterisklint@osso.nl',
        url='https://github.com/ossobv/asterisklint',
        license='GPLv3+',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            ('License :: OSI Approved :: GNU General Public License v3 '
             'or later (GPLv3+)'),
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Communications :: Telephony',
            'Topic :: Software Development :: Pre-processors',
        ],
    )

# vim: set ts=8 sw=4 sts=4 et ai tw=79:



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ..vall.builtin import register as builtin_register
from ..vall.unknown import register as unknown_register

from ..vall.app_celgenuserevent import register as app_celgenuserevent_register
from ..vall.app_chanisavail import register as app_chanisavail_register
from ..vall.app_channelredirect import register as app_channelredirect_register
from ..vall.app_chanspy import register as app_chanspy_register
from ..vall.app_confbridge import register as app_confbridge_register
from ..vall.app_dial import register as app_dial_register
from ..vall.app_directed_pickup import register as app_directed_pickup_register
from ..vall.app_dumpchan import register as app_dumpchan_register
from ..vall.app_echo import register as app_echo_register
from ..vall.app_exec import register as app_exec_register
from ..vall.app_forkcdr import register as app_forkcdr_register
from ..vall.app_macro import register as app_macro_register
from ..vall.app_meetme import register as app_meetme_register
from ..vall.app_mixmonitor import register as app_mixmonitor_register
from ..vall.app_page import register as app_page_register
from ..vall.app_playback import register as app_playback_register
from ..vall.app_queue import register as app_queue_register
from ..vall.app_read import register as app_read_register
from ..vall.app_sayunixtime import register as app_sayunixtime_register
from ..vall.app_setcallerid import register as app_setcallerid_register
from ..vall.app_stack import register as app_stack_register
from ..vall.app_system import register as app_system_register
from ..vall.app_userevent import register as app_userevent_register
from ..vall.app_verbose import register as app_verbose_register
from ..vall.app_voicemail import register as app_voicemail_register
from ..vall.app_waituntil import register as app_waituntil_register
from ..vall.chan_sip import register as chan_sip_register
from ..vall.res_fax import register as res_fax_register
from ..vall.res_indications import register as res_indications_register
from ..vall.res_musiconhold import register as res_musiconhold_register


def register(app_loader):
    for regfunc in (
            builtin_register,
            unknown_register,
            app_celgenuserevent_register,
            app_chanisavail_register,
            app_channelredirect_register,
            app_chanspy_register,
            app_confbridge_register,
            app_dial_register,
            app_directed_pickup_register,
            app_dumpchan_register,
            app_echo_register,
            app_exec_register,
            app_forkcdr_register,
            app_macro_register,
            app_meetme_register,
            app_mixmonitor_register,
            app_page_register,
            app_playback_register,
            app_queue_register,
            app_read_register,
            app_sayunixtime_register,
            app_setcallerid_register,
            app_stack_register,
            app_system_register,
            app_userevent_register,
            app_verbose_register,
            app_voicemail_register,
            app_waituntil_register,
            chan_sip_register,
            res_fax_register,
            res_indications_register,
            res_musiconhold_register):
        regfunc(app_loader)



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from tempfile import NamedTemporaryFile
import os


class FileMutatorBase(object):
    """
    Base class that mutates config files. You override this and
    implement your own process_issue method.

    Example::

        class MyMutator(FileMutatorBase):
            def process_issue(self, issue, inline, outfile):
                outfile.write(issue.changes_based_on_line(inline))

        mutator = MyMutator(
            issues_per_file={
                'somefile.conf': [
                    Issue(lineno=3, ...), Issue(lineno=5, ...)])
        mutator.print_summary()
        mutator.request_permission_and_process()
    """
    def __init__(self, issues_per_file):
        self.issues_per_file = issues_per_file
        self.filenames = list(sorted(self.issues_per_file.keys()))

    def request_permission_and_process(self):
        if not self.issues_per_file:
            print("Your dialplan has no issues. We're done here.")
            return

        print('The following files have issues:')
        print('  {}'.format('\n  '.join(self.filenames)))
        yesno = input('Update all [y/n] ? ')

        if yesno != 'y':
            print('Aborted.')
            return

        self.process()
        print('Done')

    def process(self):
        for filename in self.filenames:
            self.process_file(filename)

    def process_file(self, filename):
        # Place the tempfile in the same dir, so renaming works
        # (same device).
        tempout = NamedTemporaryFile(
            dir=os.path.dirname(filename), mode='wb', delete=False)

        issues = self.issues_per_file[filename]
        try:
            issueptr = 0
            issue = issues[issueptr]
            with open(filename, 'rb') as file_:
                for lineno, line in enumerate(file_):
                    if issue and lineno == issue.lineno:
                        # Do actual work on the issue.
                        self.process_issue(
                            issue=issue,
                            inline=line,
                            outfile=tempout)

                        # Jump to the next issue.
                        issueptr += 1
                        if issueptr < len(issues):
                            issue = issues[issueptr]
                        else:
                            issue = None
                    else:
                        tempout.write(line)
            tempout.flush()
        except:
            tempout.close()
            os.unlink(tempout.name)
            raise

        # Awesome, we've succeeded. Atomic move time!
        tempout.close()

        # Update file permissions.
        srcstat = os.stat(tempout.name)
        dststat = os.stat(filename)
        if (srcstat.st_uid != dststat.st_uid or
                srcstat.st_gid != dststat.st_gid):
            os.chown(tempout.name, dststat.st_uid, dststat.st_gid)
        if srcstat.st_mode != dststat.st_mode:
            os.chmod(tempout.name, dststat.st_mode)

        print('Overwriting', filename, '...')
        os.rename(tempout.name, filename)

    def process_issue(self, issue, inline, outfile):
        raise NotImplementedError()



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re
import os

from .configobj import EmptyLine, Context, Varset
from .configws import VerticalSpaceWarner
from .defines import ErrorDef, WarningDef, DupeDefMixin
from .file import W_WSH_BOL


if 'we_dont_want_two_linefeeds_between_classdefs':  # for flake8
    class E_CONF_BAD_LINE(ErrorDef):
        message = 'expected variable = value, line starts with {startswith!r}'

    class E_CONF_CTX_MISSING(ErrorDef):
        message = 'expected context before var/object set'

    class E_CONF_KEY_DUPE(DupeDefMixin, ErrorDef):
        message = 'duplicate varset of {key!r}'

    class E_CONF_KEY_INVALID(ErrorDef):
        message = 'expected valid var/object or new context, got {key!r}'

    class W_CONF_KEY_LATE(WarningDef):
        message = 'expected varset {key!r} earlier, fix your ordering'

    class W_CONF_CTX_DUPE(DupeDefMixin, ErrorDef):
        # BEWARE: chan_pjsip explicitly takes multiple contexts with the
        # same name. But generally it is unwanted and sometimes the
        # values are ignored or the values mask any previously set
        # values.
        message = 'duplicate context found, in some cases it is legal'

    class W_PP_EXEC(WarningDef):
        message = 'using #exec is not recommended; command is {cmd!r}'

    class W_PP_INCLUDE_MISSING(WarningDef):
        message = 'the #include file {filename!r} could not be found'

    class W_PP_QUOTE_BAD(WarningDef):
        message = 'bad quoting near preprocessor directive'

    class W_WSH_CTRL(WarningDef):
        message = 'unexpected non-space/tab whitespace or a mix'


class ProgrammingError(RuntimeError):
    def __init__(self, message):
        super().__init__(
            'An unexpected exception was raised. This is most likely a '
            'bug in the asterisklint library. If you can reproduce the '
            'problem and file an issue on the bug tracker, that would be '
            'nice. Further info: {}'.format(message))


class ConfigParser(object):
    # Parse config and emit 1=>X
    # Define who prefers "=>" and who prefers "=". Allow warnings about
    # that to be disabled.
    # Allow templating? Feature v1.1.

    # TODO: look at: main/config.c: process_text_line()
    # it will show odd escaping, and multiline stuff

    regexes = (
        # [context](template1,template2)
        (re.compile(r'^(\s*)\[([^]]*)\]\s*\(([^)\+])\)$'),
         (lambda comment, where, match: Context(
             name=match.groups()[1], templates=match.groups()[2],
             comment=comment, bolspace=match.groups()[0], where=where))),
        # [context]
        (re.compile(r'^(\s*)\[([^]]*)\]$'),
         (lambda comment, where, match: Context(
             name=match.groups()[1], templates='',
             comment=comment, bolspace=match.groups()[0], where=where))),
        # object => value
        (re.compile(r'^([^=]*?)(\s*=>\s*)(.*)$'),
         (lambda comment, where, match: Varset(
             variable=match.groups()[0], value=match.groups()[2],
             separator=match.groups()[1], comment=comment, where=where))),
        # variable = value
        (re.compile(r'^([^=]*?)(\s*=\s*)(.*)$'),
         (lambda comment, where, match: Varset(
             variable=match.groups()[0], value=match.groups()[2],
             separator=match.groups()[1], comment=comment, where=where))),
        # (void)
        (re.compile(r'^\s*$'),
         (lambda comment, where, match: EmptyLine(
             comment=comment, where=where))),
    )

    def __iter__(self):
        for where, data, comment in super(ConfigParser, self).__iter__():
            for regex, func in self.regexes:
                match = regex.match(data)
                if match:
                    # Cast the comment to a boolean, we're not using the
                    # contents, ever.
                    value = func(bool(comment), where, match)
                    yield value
                    break
            else:
                if data.lstrip().startswith('#'):
                    # Call the preprocessor, it may insert data into our
                    # internal data generator.
                    self.preprocessor(where, data, bool(comment))
                else:
                    E_CONF_BAD_LINE(where, startswith=data[0:16])

    def preprocessor(self, where, data, comment):
        text = data.lstrip()
        leading = data[0:(len(data) - len(text))]
        if leading:
            W_WSH_BOL(where)

        if text.startswith('#include'):
            cmd, rest = text[0:8], text[8:]
        elif text.startswith('#tryinclude'):
            cmd, rest = text[0:11], text[11:]
        elif text.startswith('#exec'):
            cmd, rest = text[0:5], text[5:]
        else:
            rest = None

        if not rest or ord(rest[0]) > 32:
            E_CONF_BAD_LINE(where, startswith=text[0:16])
            return

        for idx, ch in enumerate(rest):
            if ord(ch) > 32:
                break
        else:
            # Nothing after the #include/#tryinclude/#exec.
            E_CONF_BAD_LINE(where, startswith=text[0:16])
            return

        if (not all(i == ' ' for i in rest[0:idx]) and
                not all(i == '\t' for i in rest[0:idx])):
            W_WSH_CTRL(where)
        rest = rest[idx:]

        if rest[0] == '"' and rest[-1] == '"':
            rest = rest[1:-1]
        elif rest[0] == '<' and rest[-1] == '>':
            rest = rest[1:-1]
        elif (rest[0] == '<' or rest[0] == '"' or rest[-1] == '"' or
                rest[-1] == '>'):
            W_PP_QUOTE_BAD(where)

        # Very well. Let's rock.
        if cmd == '#exec':
            W_PP_EXEC(where, cmd=rest)
        elif cmd in ('#include', '#tryinclude'):
            error_if_not_exists = (cmd == '#include')
            try:
                self.include(os.path.join(os.path.dirname(where.filename),
                                          rest))
            except OSError:
                if error_if_not_exists:
                    W_PP_INCLUDE_MISSING(where, filename=rest)


class ConfigAggregator(VerticalSpaceWarner, ConfigParser):
    def __iter__(self):
        self.on_begin()

        for element in super(ConfigAggregator, self).__iter__():
            try:
                if isinstance(element, Context):
                    self.on_context(element)
                elif isinstance(element, Varset):
                    self.on_varset(element)
                elif isinstance(element, EmptyLine):
                    self.on_emptyline(element)
                else:
                    raise NotImplementedError()
            except Exception as exc:
                raise ProgrammingError(str(element.where)) from exc

        for item in self.on_yield():
            yield item

    def on_begin(self):
        self._contexts = []
        self._curcontext = None

    def on_yield(self):
        for context in self._contexts:
            yield context

    def on_context(self, context):
        # TODO: check for dupes and whatnot?
        # TODO: check for file type specific stuff?
        self._contexts.append(context)
        self._curcontext = context

    def on_varset(self, varset):
        if not self._curcontext:
            E_CONF_CTX_MISSING(varset.where)
        else:
            self._curcontext.add(varset)

    def on_emptyline(self, emptyline):
        # We don't want these.
        pass



# AsteriskLint -- an Asterisk PBX config syntax checker
# Copyright (C) 2015-2016  Walter Doekes, OSSO B.V.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from .variable import SliceMixin, Var


class ReadFunc(Var):
    """
    A special case of VarSlice where a function call is evaluated.

    The ReadFunc is different from the WriteFunc in that the ReadFunc
    has surrounding ${} tokens and allows slicing.
    """
    def __init__(self, func=None, args=None):
        assert func is not None and args is not None

        # The "convenience" strjoin in Var getslice creates trouble for
        # us here: args can be a list or a simple iterable.
        if isinstance(args, list):
            func_and_args = [func, '('] + args + [')']
        else:
            func_and_args = [func, '(', args, ')']

        super().__init__(name=Var.join(func_and_args))

        self.func = func
        self.args = args


class ReadFuncSlice(SliceMixin, ReadFunc):
    pass
